# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import tempfile
import binascii
import unittest.mock as mock
import mimetypes
from pathlib import Path

import pytest

from kserve_storage import Storage
from kserve_storage.kserve_storage import _should_download, _parse_patterns_from_env

STORAGE_MODULE = "kserve_storage.kserve_storage"
HTTPS_URI_TARGZ = "https://foo.bar/model.tar.gz"
HTTPS_URI_TARGZ_WITH_QUERY = HTTPS_URI_TARGZ + "?foo=bar"

# *.tar.gz contains a single empty file model.pth
FILE_TAR_GZ_RAW = binascii.unhexlify(
    "1f8b0800bac550600003cbcd4f49cdd12b28c960a01d3030303033315100d1e666a660dac008c28"
    "701054313a090a189919981998281a1b1b1a1118382010ddd0407a5c525894540a754656466e464e"
    "2560754969686c71ca83fe0f4281805a360140c7200009f7e1bb400060000"
)
# *.zip contains a single empty file model.pth
FILE_ZIP_RAW = binascii.unhexlify(
    "504b030414000800080035b67052000000000000000000000000090020006d6f64656c2e70746855540"
    "d000786c5506086c5506086c5506075780b000104f501000004140000000300504b07080000000002000"
    "00000000000504b0102140314000800080035b6705200000000020000000000000009002000000000000"
    "0000000a481000000006d6f64656c2e70746855540d000786c5506086c5506086c5506075780b000104f"
    "50100000414000000504b0506000000000100010057000000590000000000"
)


def test_storage_local_path():
    abs_path = "file:///"
    relative_path = "file://."
    assert Storage.download(abs_path) == abs_path.replace("file://", "", 1)
    assert Storage.download(relative_path) == relative_path.replace("file://", "", 1)


def test_storage_local_path_exception():
    not_exist_path = "file:///some/random/path"
    with pytest.raises(Exception):
        Storage.download(not_exist_path)


def test_no_prefix_local_path():
    abs_path = "/"
    relative_path = "."
    assert Storage.download(abs_path) == abs_path
    assert Storage.download(relative_path) == relative_path


def test_local_path_with_out_dir_exist():
    abs_path = "file:///tmp"
    out_dir = "/tmp"
    assert Storage.download(abs_path, out_dir=out_dir) == out_dir


def test_local_path_with_out_dir_not_exist():
    abs_path = "file:///tmp"
    out_dir = "/tmp/test-abc"
    assert Storage.download(abs_path, out_dir=out_dir) == out_dir


class MockHttpResponse(object):
    def __init__(self, status_code=404, raw=b"", content_type=""):
        self.status_code = status_code
        self.raw = io.BytesIO(raw)
        self.headers = {"Content-Type": content_type}

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_val, traceback):
        pass


@mock.patch(
    "requests.get",
    return_value=MockHttpResponse(
        status_code=200, content_type="application/octet-stream"
    ),
)
def test_http_uri_path(_):
    http_uri = "http://foo.bar/model.joblib"
    http_with_query_uri = "http://foo.bar/model.joblib?foo=bar"
    out_dir = "."
    assert Storage.download(http_uri, out_dir=out_dir) == out_dir
    assert Storage.download(http_with_query_uri, out_dir=out_dir) == out_dir
    os.remove("./model.joblib")


@mock.patch(
    "requests.get",
    return_value=MockHttpResponse(
        status_code=200, content_type="application/octet-stream"
    ),
)
def test_https_uri_path(_):
    https_uri = "https://foo.bar/model.joblib"
    https_with_query_uri = "https://foo.bar/model.joblib?foo=bar"
    out_dir = "."
    assert Storage.download(https_uri, out_dir=out_dir) == out_dir
    assert Storage.download(https_with_query_uri, out_dir=out_dir) == out_dir
    os.remove("./model.joblib")


http_uri_path_testparams = [
    (
        HTTPS_URI_TARGZ,
        MockHttpResponse(200, FILE_TAR_GZ_RAW, "application/x-tar"),
        None,
    ),
    (
        HTTPS_URI_TARGZ,
        MockHttpResponse(200, FILE_TAR_GZ_RAW, "application/x-gtar"),
        None,
    ),
    (
        HTTPS_URI_TARGZ,
        MockHttpResponse(200, FILE_TAR_GZ_RAW, "application/x-gzip"),
        None,
    ),
    (HTTPS_URI_TARGZ, MockHttpResponse(200, FILE_TAR_GZ_RAW, "application/gzip"), None),
    (
        HTTPS_URI_TARGZ,
        MockHttpResponse(200, FILE_TAR_GZ_RAW, "application/zip"),
        RuntimeError,
    ),
    (
        HTTPS_URI_TARGZ_WITH_QUERY,
        MockHttpResponse(200, FILE_TAR_GZ_RAW, "application/x-tar"),
        None,
    ),
    (
        HTTPS_URI_TARGZ_WITH_QUERY,
        MockHttpResponse(200, FILE_TAR_GZ_RAW, "application/x-gtar"),
        None,
    ),
    (
        HTTPS_URI_TARGZ_WITH_QUERY,
        MockHttpResponse(200, FILE_TAR_GZ_RAW, "application/x-gzip"),
        None,
    ),
    (
        HTTPS_URI_TARGZ_WITH_QUERY,
        MockHttpResponse(200, FILE_TAR_GZ_RAW, "application/gzip"),
        None,
    ),
    (
        "https://foo.bar/model.zip",
        MockHttpResponse(200, FILE_ZIP_RAW, "application/zip"),
        None,
    ),
    (
        "https://foo.bar/model.zip",
        MockHttpResponse(200, FILE_ZIP_RAW, "application/x-zip-compressed"),
        None,
    ),
    (
        "https://foo.bar/model.zip",
        MockHttpResponse(200, FILE_ZIP_RAW, "application/zip-compressed"),
        None,
    ),
    (
        "https://foo.bar/model.zip?foo=bar",
        MockHttpResponse(200, FILE_ZIP_RAW, "application/zip"),
        None,
    ),
    (
        "https://foo.bar/model.zip?foo=bar",
        MockHttpResponse(200, FILE_ZIP_RAW, "application/x-zip-compressed"),
        None,
    ),
    (
        "https://foo.bar/model.zip?foo=bar",
        MockHttpResponse(200, FILE_ZIP_RAW, "application/zip-compressed"),
        None,
    ),
    ("https://theabyss.net/model.joblib", MockHttpResponse(404), RuntimeError),
    (
        "https://some.site.com/test.model",
        MockHttpResponse(status_code=200, content_type="text/html"),
        RuntimeError,
    ),
    ("https://foo.bar/test/", MockHttpResponse(200), RuntimeError),
    (
        "https://foo.bar/download?path=/20210530/model.zip",
        MockHttpResponse(200, FILE_ZIP_RAW, "application/zip"),
        None,
    ),
    (
        "https://foo.bar/download?path=/20210530/model.zip",
        MockHttpResponse(200, FILE_ZIP_RAW, "application/x-zip" "-compressed"),
        None,
    ),
    (
        "https://foo.bar/download?path=/20210530/model.zip",
        MockHttpResponse(200, FILE_ZIP_RAW, "application/zip" "-compressed"),
        None,
    ),
]


@pytest.mark.parametrize("uri,response,expected_error", http_uri_path_testparams)
def test_http_uri_paths(uri, response, expected_error):
    if expected_error:

        def test(_):
            with pytest.raises(expected_error):
                Storage.download(uri)

    else:

        def test(_):
            with tempfile.TemporaryDirectory() as out_dir:
                assert Storage.download(uri, out_dir=out_dir) == out_dir
                assert os.path.exists(os.path.join(out_dir, "model.pth"))

    mock.patch("requests.get", return_value=response)(test)()


def test_storage_blob_exception():
    blob_path = "https://localhost:1/container/some/blob/"
    with pytest.raises(Exception):
        Storage.download(blob_path)


def test_unpack_tar_file():
    out_dir = "."
    tar_file = os.path.join(out_dir, "model.tgz")
    Path(tar_file).write_bytes(FILE_TAR_GZ_RAW)
    mimetype, _ = mimetypes.guess_type(tar_file)
    Storage._unpack_archive_file(tar_file, mimetype, out_dir)
    assert os.path.exists(os.path.join(out_dir, "model.pth"))
    os.remove(os.path.join(out_dir, "model.pth"))


def test_unpack_zip_file():
    out_dir = "."
    tar_file = os.path.join(out_dir, "model.zip")
    Path(tar_file).write_bytes(FILE_ZIP_RAW)
    mimetype, _ = mimetypes.guess_type(tar_file)
    Storage._unpack_archive_file(tar_file, mimetype, out_dir)
    assert os.path.exists(os.path.join(out_dir, "model.pth"))
    os.remove(os.path.join(out_dir, "model.pth"))


@mock.patch(STORAGE_MODULE + ".Storage._download_azure_blob")
def test_download_azure_blob_called_with_matching_uri(mock_download_azure_blob):
    azure_blob_uris = [
        "https://accountname.blob.core.windows.net/container/some/blob/",
        "https://accountname.z20.blob.storage.azure.net/container/some/blob/",
        "https://accountname.z2.blob.storage.azure.net/container/some/blob/",
    ]

    for uri in azure_blob_uris:
        Storage.download(uri, out_dir="dest_path")

    expected_calls = [
        mock.call(uri, "dest_path", None, None) for uri in azure_blob_uris
    ]
    mock_download_azure_blob.assert_has_calls(expected_calls)


@mock.patch(STORAGE_MODULE + ".Storage._download_azure_file_share")
def test_download_azure_file_share_called_with_matching_uri(
    mock_download_azure_file_share,
):
    azure_file_uris = [
        "https://accountname.file.core.windows.net/container/some/blob",
        "https://accountname.z20.file.storage.azure.net/container/some/blob",
        "https://accountname.z2.file.storage.azure.net/container/some/blob",
    ]

    for uri in azure_file_uris:
        Storage.download(uri, out_dir="dest_path")

    expected_calls = [
        mock.call(uri, "dest_path", None, None) for uri in azure_file_uris
    ]
    mock_download_azure_file_share.assert_has_calls(expected_calls)


git_repo_test_params = [
    # (uri, username_in_url, username_env, password_env, expected_clean_uri)
    (
        "https://github.com/user/repo.git",
        None,
        None,
        None,
        "https://github.com/user/repo.git",
    ),
    (
        "https://username@github.com/user/repo.git",
        "username",
        None,
        None,
        "https://github.com/user/repo.git",
    ),
    (
        "https://github.com/user/repo.git",
        None,
        "env_username",
        "env_password",
        "https://github.com/user/repo.git",
    ),
    (
        "https://username@github.com/user/repo.git",
        "username",
        None,
        "env_password",
        "https://github.com/user/repo.git",
    ),
]


@pytest.mark.parametrize(
    "uri,username_in_url,username_env,password_env,expected_clean_uri",
    git_repo_test_params,
)
@mock.patch("dulwich.porcelain.clone")
def test_git_repo_download_success(
    mock_clone, uri, username_in_url, username_env, password_env, expected_clean_uri
):
    """Test successful git repository downloads with HTTPS authentication."""
    out_dir = "/tmp/test_model"

    env_vars = {}
    if username_env:
        env_vars["GIT_USERNAME"] = username_env
    if password_env:
        env_vars["GIT_PASSWORD"] = password_env

    with mock.patch.dict(os.environ, env_vars):
        result = Storage.download(uri, out_dir=out_dir)

    assert result == out_dir

    # Verify dulwich.porcelain.clone was called with correct arguments
    mock_clone.assert_called_once()
    call_args = mock_clone.call_args

    # Check URI (should be clean URI without username)
    assert call_args[0][0] == expected_clean_uri
    assert call_args[0][1] == out_dir

    # Check keyword arguments
    kwargs = call_args[1]
    assert kwargs["depth"] == 1

    # Check username (from URL or env var)
    expected_username = username_in_url or username_env
    if expected_username:
        assert kwargs["username"] == expected_username
    else:
        assert "username" not in kwargs

    # Check password (from env var)
    if password_env:
        assert kwargs["password"] == password_env
    else:
        assert "password" not in kwargs


@mock.patch("dulwich.porcelain.clone")
def test_git_repo_download_git_protocol_error(mock_clone):
    from dulwich.errors import GitProtocolError

    uri = "https://github.com/user/nonexistent.git"
    out_dir = "/tmp/test_model"

    mock_clone.side_effect = GitProtocolError("Authentication failed")

    with pytest.raises(RuntimeError) as exc_info:
        Storage.download(uri, out_dir=out_dir)

    # Unified error handling produces user-friendly messages
    assert "Git" in str(exc_info.value)
    assert "authentication" in str(exc_info.value).lower()


git_error_test_params = [
    (Exception("Repository not found"), "not found"),
    (Exception("Authentication failed"), "authentication"),
    (Exception("Network error"), "git"),
]


@pytest.mark.parametrize("exception,expected_message", git_error_test_params)
@mock.patch("dulwich.porcelain.clone")
def test_git_repo_download_errors(mock_clone, exception, expected_message):
    uri = "https://github.com/user/nonexistent.git"
    out_dir = "/tmp/test_model"

    mock_clone.side_effect = exception

    with pytest.raises(RuntimeError) as exc_info:
        Storage.download(uri, out_dir=out_dir)

    # Unified error handling produces user-friendly messages
    assert expected_message in str(exc_info.value).lower()


@mock.patch("dulwich.porcelain.clone")
def test_git_repo_download_public_repo_no_auth(mock_clone):
    uri = "https://github.com/user/public-repo.git"
    out_dir = "/tmp/test_model"

    result = Storage.download(uri, out_dir=out_dir)

    assert result == out_dir
    mock_clone.assert_called_once()
    call_args = mock_clone.call_args
    kwargs = call_args[1]
    assert kwargs["depth"] == 1
    # No username or password should be passed for public repos
    assert "username" not in kwargs
    assert "password" not in kwargs


# Tests for _should_download and _parse_patterns_from_env


class TestShouldDownload:
    def test_no_patterns_allows_all(self):
        assert _should_download("model.safetensors") is True
        assert _should_download("subdir/model.bin") is True

    def test_allow_patterns_match(self):
        assert (
            _should_download("model.safetensors", allow_patterns=["*.safetensors"])
            is True
        )
        assert _should_download("model.bin", allow_patterns=["*.safetensors"]) is False

    def test_allow_patterns_multiple(self):
        patterns = ["*.safetensors", "*.json"]
        assert _should_download("model.safetensors", allow_patterns=patterns) is True
        assert _should_download("config.json", allow_patterns=patterns) is True
        assert _should_download("model.bin", allow_patterns=patterns) is False

    def test_ignore_patterns_match(self):
        assert _should_download("model.safetensors", ignore_patterns=["*.bin"]) is True
        assert _should_download("model.bin", ignore_patterns=["*.bin"]) is False

    def test_ignore_patterns_multiple(self):
        patterns = ["*.bin", "*.gguf"]
        assert _should_download("model.safetensors", ignore_patterns=patterns) is True
        assert _should_download("model.bin", ignore_patterns=patterns) is False
        assert _should_download("model.gguf", ignore_patterns=patterns) is False

    def test_both_patterns(self):
        allow = ["*.safetensors", "*.json"]
        ignore = ["config.json"]
        assert _should_download("model.safetensors", allow, ignore) is True
        assert _should_download("config.json", allow, ignore) is False
        assert _should_download("tokenizer.json", allow, ignore) is True
        assert _should_download("model.bin", allow, ignore) is False

    def test_subdirectory_matching(self):
        assert (
            _should_download(
                "subdir/model.safetensors", allow_patterns=["*.safetensors"]
            )
            is True
        )
        assert (
            _should_download("deep/nested/model.bin", ignore_patterns=["*.bin"])
            is False
        )

    def test_empty_patterns_treated_as_none(self):
        assert _should_download("model.bin", allow_patterns=[]) is True
        assert _should_download("model.bin", ignore_patterns=[]) is True


class TestParsePatterns:
    def test_not_set(self):
        with mock.patch.dict(os.environ, {}, clear=True):
            assert _parse_patterns_from_env("STORAGE_ALLOW_PATTERNS") is None

    def test_empty_string(self):
        with mock.patch.dict(os.environ, {"STORAGE_ALLOW_PATTERNS": ""}):
            assert _parse_patterns_from_env("STORAGE_ALLOW_PATTERNS") is None

    def test_json_array(self):
        with mock.patch.dict(
            os.environ,
            {"STORAGE_ALLOW_PATTERNS": '["*.safetensors", "*.json"]'},
        ):
            result = _parse_patterns_from_env("STORAGE_ALLOW_PATTERNS")
            assert result == ["*.safetensors", "*.json"]

    def test_comma_separated(self):
        with mock.patch.dict(
            os.environ, {"STORAGE_ALLOW_PATTERNS": "*.safetensors,*.json"}
        ):
            result = _parse_patterns_from_env("STORAGE_ALLOW_PATTERNS")
            assert result == ["*.safetensors", "*.json"]

    def test_single_pattern(self):
        with mock.patch.dict(os.environ, {"STORAGE_ALLOW_PATTERNS": "*.safetensors"}):
            result = _parse_patterns_from_env("STORAGE_ALLOW_PATTERNS")
            assert result == ["*.safetensors"]

    def test_json_single_string(self):
        with mock.patch.dict(os.environ, {"STORAGE_ALLOW_PATTERNS": '"*.safetensors"'}):
            result = _parse_patterns_from_env("STORAGE_ALLOW_PATTERNS")
            assert result == ["*.safetensors"]



# Copyright 2024 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Storage error handling utilities.

This module provides unified error handling for all storage protocols,
converting low-level exceptions into user-friendly error messages.
"""

from .logging import logger


def raise_storage_error(
    protocol: str, uri: str, error: Exception, resource_name: str = ""
) -> None:
    """
    Unified error handler for all storage protocols.
    Logs the error and raises RuntimeError with a user-friendly message.

    Args:
        protocol: The storage protocol (e.g., "S3", "GCS", "Azure")
        uri: The storage URI being accessed
        error: The original exception
        resource_name: Optional resource name for context (e.g., bucket name)

    Raises:
        RuntimeError: With a user-friendly error message
    """
    error_msg = get_storage_error_message(protocol, error, resource_name)
    logger.error("%s error for %s: %s", protocol, uri, error)
    raise RuntimeError(error_msg) from error


def get_storage_error_message(
    protocol: str, error: Exception, resource_name: str = ""
) -> str:
    """
    Get user-friendly error message for storage errors across all protocols.

    Args:
        protocol: The storage protocol (e.g., "S3", "GCS", "Azure")
        error: The original exception
        resource_name: Optional resource name for context

    Returns:
        A user-friendly error message string
    """
    error_type = type(error).__name__
    error_str = str(error).lower()

    # Authentication errors
    auth_error_types = (
        "NoCredentialsError",
        "DefaultCredentialsError",
        "GoogleAuthError",
        "ClientAuthenticationError",
        "GatedRepoError",
    )
    if (
        error_type in auth_error_types
        or "credential" in error_str
        or "auth" in error_str
    ):
        auth_hints = {
            "S3": "Set AWS_ACCESS_KEY_ID/AWS_SECRET_ACCESS_KEY or use awsAnonymousCredential=true.",
            "GCS": "Verify GOOGLE_APPLICATION_CREDENTIALS or use anonymous access.",
            "Azure": "Verify AZURE_CLIENT_ID/AZURE_CLIENT_SECRET or AZURE_STORAGE_ACCESS_KEY.",
            "HDFS": "Verify Kerberos keytab and principal configuration.",
            "HuggingFace": "Set HF_TOKEN or request access to the gated repository.",
            "HTTP": "Verify authentication credentials.",
            "Git": "Set GIT_USERNAME/GIT_PASSWORD or use a public repository.",
        }
        return "%s authentication failed. %s" % (
            protocol,
            auth_hints.get(protocol, ""),
        )

    # Not found errors
    not_found_types = (
        "NotFound",
        "ResourceNotFoundError",
        "RepositoryNotFoundError",
        "RevisionNotFoundError",
    )
    if (
        error_type in not_found_types
        or "not found" in error_str
        or "does not exist" in error_str
    ):
        if resource_name:
            return "%s resource '%s' not found." % (protocol, resource_name)
        return "%s resource not found." % protocol

    # Access denied errors
    access_denied_types = ("Forbidden", "AccessDenied")
    if (
        error_type in access_denied_types
        or "access denied" in error_str
        or "permission" in error_str
    ):
        if resource_name:
            return "Access denied to %s resource '%s'. Verify permissions." % (
                protocol,
                resource_name,
            )
        return "Access denied. Verify %s permissions." % protocol

    # Connection errors
    if (
        error_type == "ConnectionError"
        or "connection" in error_str
        or "timeout" in error_str
    ):
        if resource_name:
            return "Failed to connect to %s endpoint '%s'." % (
                protocol,
                resource_name,
            )
        return "Failed to connect to %s endpoint." % protocol

    # S3-specific error codes
    if protocol == "S3" and hasattr(error, "response"):
        error_code = error.response.get("Error", {}).get("Code", "")
        if error_code in ("InvalidAccessKeyId", "SignatureDoesNotMatch"):
            return "Invalid AWS credentials. Verify AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY."
        if error_code == "NoSuchBucket":
            return "S3 bucket '%s' does not exist." % resource_name
        if error_code:
            return "S3 error [%s]: %s" % (
                error_code,
                error.response.get("Error", {}).get("Message", ""),
            )

    # Default: include error type and message
    return "%s error: %s" % (protocol, error)


def check_http_response(uri: str, response) -> None:
    """
    Check HTTP response status and raise RuntimeError for error codes.

    Args:
        uri: The HTTP URI being accessed
        response: The HTTP response object

    Raises:
        RuntimeError: If the response indicates an error
    """
    status_messages = {
        401: "HTTP authentication failed for '%s'. Verify credentials.",
        403: "HTTP access denied for '%s'. Verify permissions.",
        404: "HTTP resource not found at '%s'.",
    }
    if response.status_code in status_messages:
        raise RuntimeError(status_messages[response.status_code] % uri)
    if response.status_code != 200:
        raise RuntimeError(
            "HTTP request to '%s' failed with status code %s."
            % (uri, response.status_code)
        )



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest.mock as mock
import pytest
import shutil

from kserve_storage import Storage

STORAGE_MODULE = "kserve_storage.kserve_storage"


def create_mock_item(path):
    mock_obj = mock.MagicMock()
    mock_obj.name = path
    mock_obj.size = 1024  # Mock non-zero size for regular files

    # For async streaming, we need to mock the chunks() method
    async def mock_chunks():
        yield b"test"

    mock_stream = mock.MagicMock()
    mock_stream.chunks = mock_chunks
    mock_obj.readall.return_value = b"test"
    mock_obj.chunks = mock_chunks
    return mock_obj


def create_mock_blob(mock_storage, paths):
    mock_objs = [create_mock_item(path) for path in paths]
    mock_container = mock.MagicMock()

    # Mock async list_blobs (changed from walk_blobs)
    async def mock_list_blobs(name_starts_with=None, include=None):
        for obj in mock_objs:
            yield obj

    mock_container.list_blobs = mock_list_blobs

    # Mock async download_blob returning a stream (no blob_name param needed)
    async def mock_download_blob(max_concurrency=None):
        # Return a stream-like object with chunks() method
        stream = mock.MagicMock()

        async def mock_chunks():
            yield b"test"

        stream.chunks = mock_chunks
        return stream

    mock_blob_client = mock.MagicMock()
    mock_blob_client.download_blob = mock_download_blob
    mock_container.get_blob_client.return_value = mock_blob_client

    mock_svc = mock.MagicMock()
    mock_svc.get_container_client.return_value = mock_container

    # Make the service client work as an async context manager
    async def mock_aenter(self):
        return mock_svc

    async def mock_aexit(self, exc_type, exc_val, exc_tb):
        return None

    mock_svc.__aenter__ = mock_aenter
    mock_svc.__aexit__ = mock_aexit

    mock_storage.return_value = mock_svc
    return mock_storage, mock_container


def create_mock_dir(name):
    mock_file = mock.MagicMock()
    mock_file.name = name
    mock_file.is_directory = True
    return mock_file


def create_mock_file(name):
    mock_file = mock.MagicMock()
    mock_file.name = name
    mock_file.is_directory = False
    return mock_file


def create_mock_objects_for_file_share(mock_storage, mock_file_items):
    mock_share = mock.MagicMock()
    mock_share.list_directories_and_files.side_effect = mock_file_items

    mock_file = mock.MagicMock()
    mock_share.get_file_client.return_value = mock_file
    mock_data = mock.MagicMock()
    mock_file.download_file.return_value = mock_data

    mock_svc = mock.MagicMock()
    mock_svc.get_share_client.return_value = mock_share

    mock_storage.return_value = mock_svc
    return mock_storage, mock_share, mock_data


def get_call_args(call_args_list):
    arg_list = []
    for call in call_args_list:
        args, _ = call
        arg_list.append(args)
    return arg_list


# pylint: disable=protected-access
@pytest.fixture(scope="session", autouse=True)
def test_cleanup():
    yield None
    # Will be executed after the last test
    shutil.rmtree("some", ignore_errors=True)
    shutil.rmtree("dest_path", ignore_errors=True)


@mock.patch(STORAGE_MODULE + ".os.makedirs")
@mock.patch("azure.storage.blob.aio.BlobServiceClient")
def test_blob(mock_storage, mock_makedirs):  # pylint: disable=unused-argument

    # given
    blob_path = "https://kfserving.blob.core.windows.net/triton/simple_string/"
    paths = ["simple_string/1/model.graphdef", "simple_string/config.pbtxt"]
    mock_blob, mock_container = create_mock_blob(mock_storage, paths)

    # when
    Storage._download_azure_blob(blob_path, "dest_path")

    # then
    # Check that get_blob_client was called for each blob
    arg_list = get_call_args(mock_container.get_blob_client.call_args_list)
    assert set(arg_list) == set(
        [("simple_string/1/model.graphdef",), ("simple_string/config.pbtxt",)]
    )

    mock_storage.assert_called_with(
        "https://kfserving.blob.core.windows.net", credential=None
    )


@mock.patch(STORAGE_MODULE + ".os.makedirs")
@mock.patch("azure.storage.blob.aio.BlobServiceClient")
def test_blob_file_direct(
    mock_storage, mock_makedirs
):  # pylint: disable=unused-argument

    # given
    blob_path = "https://accountname.blob.core.windows.net/container/somefile.text"
    paths = ["somefile.text"]
    mock_blob, mock_container = create_mock_blob(mock_storage, paths)

    # when
    Storage._download_azure_blob(blob_path, "dest_path")

    # then
    arg_list = get_call_args(mock_container.get_blob_client.call_args_list)
    assert arg_list == [("somefile.text",)]
    mock_storage.assert_called_with(
        "https://accountname.blob.core.windows.net", credential=None
    )


@mock.patch(STORAGE_MODULE + ".os.makedirs")
@mock.patch(STORAGE_MODULE + ".Storage._get_azure_storage_token")
@mock.patch("azure.storage.blob.aio.BlobServiceClient")
def test_secure_blob(
    mock_storage, mock_get_token, mock_makedirs
):  # pylint: disable=unused-argument

    # given
    blob_path = "https://kfsecured.blob.core.windows.net/triton/simple_string/"
    mock_get_token.return_value = "some_token"
    # Set up empty blob mock to trigger RuntimeError
    mock_blob, mock_container = create_mock_blob(mock_storage, [])

    # when
    with pytest.raises(RuntimeError):
        Storage._download_azure_blob(blob_path, "dest_path")

    # then
    mock_get_token.assert_called()
    arg_list = []
    for call in mock_storage.call_args_list:
        _, kwargs = call
        arg_list.append(kwargs)
    assert arg_list == [{"credential": "some_token"}]


@mock.patch(STORAGE_MODULE + ".os.makedirs")
@mock.patch("azure.storage.blob.aio.BlobServiceClient")
def test_deep_blob(mock_storage, mock_makedirs):  # pylint: disable=unused-argument

    # given
    blob_path = (
        "https://accountname.blob.core.windows.net/container/some/deep/blob/path"
    )
    paths = ["f1", "f2", "d1/f11", "d1/d2/f21", "d1/d2/d3/f1231", "d4/f41"]
    fq_item_paths = ["some/deep/blob/path/" + p for p in paths]
    expected_calls = [(f,) for f in fq_item_paths]

    # when
    mock_blob, mock_container = create_mock_blob(mock_storage, fq_item_paths)
    try:
        Storage._download_azure_blob(blob_path, "some/dest/path")
    except OSError:  # Permissions Error Handling
        pass

    # then
    actual_calls = get_call_args(mock_container.get_blob_client.call_args_list)
    assert set(actual_calls) == set(expected_calls)


@mock.patch(STORAGE_MODULE + ".os.makedirs")
@mock.patch("azure.storage.blob.aio.BlobServiceClient")
def test_blob_file(mock_storage, mock_makedirs):  # pylint: disable=unused-argument

    # given
    blob_path = "https://accountname.blob.core.windows.net/container/somefile.text"
    paths = ["somefile"]
    fq_item_paths = paths
    expected_calls = [(f,) for f in fq_item_paths]

    # when
    mock_blob, mock_container = create_mock_blob(mock_storage, paths)
    Storage._download_azure_blob(blob_path, "some/dest/path")

    # then
    actual_calls = get_call_args(mock_container.get_blob_client.call_args_list)
    assert actual_calls == expected_calls


@mock.patch(STORAGE_MODULE + ".os.makedirs")
@mock.patch("azure.storage.blob.aio.BlobServiceClient")
def test_blob_fq_file(mock_storage, mock_makedirs):  # pylint: disable=unused-argument

    # given
    blob_path = (
        "https://accountname.blob.core.windows.net/container/folder/somefile.text"
    )
    paths = ["somefile"]
    fq_item_paths = ["folder/" + p for p in paths]
    expected_calls = [(f,) for f in fq_item_paths]

    # when
    mock_blob, mock_container = create_mock_blob(mock_storage, fq_item_paths)
    Storage._download_azure_blob(blob_path, "some/dest/path")

    # then
    actual_calls = get_call_args(mock_container.get_blob_client.call_args_list)
    assert actual_calls == expected_calls


@mock.patch(STORAGE_MODULE + ".os.makedirs")
@mock.patch("azure.storage.blob.aio.BlobServiceClient")
def test_blob_no_prefix(mock_storage, mock_makedirs):  # pylint: disable=unused-argument

    # given
    blob_path = "https://accountname.blob.core.windows.net/container/"
    paths = ["somefile.text", "somefolder/somefile.text"]
    fq_item_paths = ["" + p for p in paths]
    expected_calls = [(f,) for f in fq_item_paths]

    # when
    mock_blob, mock_container = create_mock_blob(mock_storage, fq_item_paths)
    Storage._download_azure_blob(blob_path, "some/dest/path")

    # then
    actual_calls = get_call_args(mock_container.get_blob_client.call_args_list)
    assert set(actual_calls) == set(expected_calls)


@mock.patch(STORAGE_MODULE + ".os.makedirs")
@mock.patch(STORAGE_MODULE + ".Storage._get_azure_storage_access_key")
@mock.patch("azure.storage.fileshare.ShareServiceClient")
def test_file_share(
    mock_storage, mock_get_access_key, mock_makedirs
):  # pylint: disable=unused-argument

    # given
    file_share_path = "https://kfserving.file.core.windows.net/triton/simple_string/"
    mock_get_access_key.return_value = "some_token"

    mock_file_share, mock_file, mock_data = create_mock_objects_for_file_share(
        mock_storage,
        [
            [create_mock_dir("1"), create_mock_file("config.pbtxt")],
            [create_mock_file("model.graphdef")],
            [],
        ],
    )

    # when
    Storage._download_azure_file_share(file_share_path, "dest_path")

    # then
    arg_list = get_call_args(mock_file.get_file_client.call_args_list)
    assert set(arg_list) == set(
        [("simple_string/1/model.graphdef",), ("simple_string/config.pbtxt",)]
    )

    # then
    mock_get_access_key.assert_called()
    mock_storage.assert_called_with(
        "https://kfserving.file.core.windows.net", credential="some_token"
    )


@mock.patch(STORAGE_MODULE + ".os.makedirs")
@mock.patch(STORAGE_MODULE + ".Storage._get_azure_storage_access_key")
@mock.patch("azure.storage.fileshare.ShareServiceClient")
def test_deep_file_share(
    mock_storage, mock_get_access_key, mock_makedirs
):  # pylint: disable=unused-argument

    file_share_path = (
        "https://accountname.file.core.windows.net/container/some/deep/blob/path"
    )
    paths = ["f1", "f2", "d1/f11", "d1/d2/f21", "d1/d2/d3/f1231", "d4/f41"]
    fq_item_paths = ["some/deep/blob/path/" + p for p in paths]
    expected_calls = [(f,) for f in fq_item_paths]
    mock_get_access_key.return_value = "some_token"

    # when
    mock_file_share, mock_file, mock_data = create_mock_objects_for_file_share(
        mock_storage,
        [
            [
                create_mock_dir("d1"),
                create_mock_dir("d4"),
                create_mock_file("f1"),
                create_mock_file("f2"),
            ],
            [create_mock_file("f41")],
            [create_mock_dir("d2"), create_mock_file("f11")],
            [create_mock_dir("d3"), create_mock_file("f21")],
            [create_mock_file("f1231")],
            [],
        ],
    )
    try:
        Storage._download_azure_file_share(file_share_path, "some/dest/path")
    except OSError:  # Permissions Error Handling
        pass

    # then
    actual_calls = get_call_args(mock_file.get_file_client.call_args_list)
    assert set(actual_calls) == set(expected_calls)

    # then
    mock_get_access_key.assert_called()
    mock_storage.assert_called_with(
        "https://accountname.file.core.windows.net", credential="some_token"
    )


@mock.patch(STORAGE_MODULE + ".os.makedirs")
@mock.patch(STORAGE_MODULE + ".Storage._get_azure_storage_access_key")
@mock.patch("azure.storage.fileshare.ShareServiceClient")
def test_file_share_fq_file(
    mock_storage, mock_get_access_key, mock_makedirs
):  # pylint: disable=unused-argument

    # given
    file_share_path = "https://accountname.file.core.windows.net/container/folder/"
    paths = ["somefile.text"]
    fq_item_paths = ["folder/" + p for p in paths]
    expected_calls = [(f,) for f in fq_item_paths]

    # when
    mock_get_access_key.return_value = "some_token"
    mock_file_share, mock_file, mock_data = create_mock_objects_for_file_share(
        mock_storage, [[create_mock_file("somefile.text")], []]
    )
    Storage._download_azure_file_share(file_share_path, "some/dest/path")

    # then
    actual_calls = get_call_args(mock_file.get_file_client.call_args_list)
    assert actual_calls == expected_calls

    # then
    mock_get_access_key.assert_called()
    mock_storage.assert_called_with(
        "https://accountname.file.core.windows.net", credential="some_token"
    )


@mock.patch(STORAGE_MODULE + ".os.makedirs")
@mock.patch(STORAGE_MODULE + ".Storage._get_azure_storage_access_key")
@mock.patch("azure.storage.fileshare.ShareServiceClient")
def test_file_share_no_prefix(
    mock_storage, mock_get_access_key, mock_makedirs
):  # pylint: disable=unused-argument

    # given
    file_share_path = "https://accountname.file.core.windows.net/container/"
    paths = ["somefile.text", "somefolder/somefile.text"]
    fq_item_paths = ["" + p for p in paths]
    expected_calls = [(f,) for f in fq_item_paths]

    # when
    mock_get_access_key.return_value = "some_token"
    mock_file_share, mock_file, mock_data = create_mock_objects_for_file_share(
        mock_storage,
        [
            [create_mock_dir("somefolder"), create_mock_file("somefile.text")],
            [create_mock_file("somefile.text")],
            [],
        ],
    )
    Storage._download_azure_file_share(file_share_path, "some/dest/path")

    # then
    arg_list = get_call_args(mock_file.get_file_client.call_args_list)
    assert set(arg_list) == set(expected_calls)

    # then
    mock_get_access_key.assert_called()
    mock_storage.assert_called_with(
        "https://accountname.file.core.windows.net", credential="some_token"
    )


@mock.patch(STORAGE_MODULE + ".os.makedirs")
@mock.patch("azure.storage.blob.aio.BlobServiceClient")
def test_blob_allow_patterns(mock_storage, mock_makedirs):
    # given
    blob_path = "https://kfserving.blob.core.windows.net/triton/model/"
    paths = [
        "model/weights.safetensors",
        "model/weights.bin",
        "model/config.json",
    ]
    mock_blob, mock_container = create_mock_blob(mock_storage, paths)

    # when
    Storage._download_azure_blob(
        blob_path, "dest_path", allow_patterns=["*.safetensors", "*.json"]
    )

    # then
    arg_list = get_call_args(mock_container.get_blob_client.call_args_list)
    blob_names = [a[0] for a in arg_list]
    assert "model/weights.safetensors" in blob_names
    assert "model/config.json" in blob_names
    assert "model/weights.bin" not in blob_names


@mock.patch(STORAGE_MODULE + ".os.makedirs")
@mock.patch("azure.storage.blob.aio.BlobServiceClient")
def test_blob_ignore_patterns(mock_storage, mock_makedirs):
    # given
    blob_path = "https://kfserving.blob.core.windows.net/triton/model/"
    paths = [
        "model/weights.safetensors",
        "model/weights.bin",
        "model/config.json",
    ]
    mock_blob, mock_container = create_mock_blob(mock_storage, paths)

    # when
    Storage._download_azure_blob(blob_path, "dest_path", ignore_patterns=["*.bin"])

    # then
    arg_list = get_call_args(mock_container.get_blob_client.call_args_list)
    blob_names = [a[0] for a in arg_list]
    assert "model/weights.bin" not in blob_names
    assert len(blob_names) == 2


@mock.patch(STORAGE_MODULE + ".os.makedirs")
@mock.patch(STORAGE_MODULE + ".Storage._get_azure_storage_access_key")
@mock.patch("azure.storage.fileshare.ShareServiceClient")
def test_file_share_allow_patterns(mock_storage, mock_get_access_key, mock_makedirs):
    # given
    file_share_path = "https://kfserving.file.core.windows.net/triton/model/"
    mock_get_access_key.return_value = "some_token"

    mock_file_share, mock_file, mock_data = create_mock_objects_for_file_share(
        mock_storage,
        [
            [
                create_mock_file("weights.safetensors"),
                create_mock_file("weights.bin"),
                create_mock_file("config.json"),
            ],
            [],
        ],
    )

    # when
    Storage._download_azure_file_share(
        file_share_path, "dest_path", allow_patterns=["*.safetensors", "*.json"]
    )

    # then
    arg_list = get_call_args(mock_file.get_file_client.call_args_list)
    file_paths = [a[0] for a in arg_list]
    assert "model/weights.safetensors" in file_paths
    assert "model/config.json" in file_paths
    assert "model/weights.bin" not in file_paths


@mock.patch(STORAGE_MODULE + ".os.makedirs")
@mock.patch(STORAGE_MODULE + ".Storage._get_azure_storage_access_key")
@mock.patch("azure.storage.fileshare.ShareServiceClient")
def test_file_share_ignore_patterns(mock_storage, mock_get_access_key, mock_makedirs):
    # given
    file_share_path = "https://kfserving.file.core.windows.net/triton/model/"
    mock_get_access_key.return_value = "some_token"

    mock_file_share, mock_file, mock_data = create_mock_objects_for_file_share(
        mock_storage,
        [
            [
                create_mock_file("weights.safetensors"),
                create_mock_file("weights.bin"),
                create_mock_file("config.json"),
            ],
            [],
        ],
    )

    # when
    Storage._download_azure_file_share(
        file_share_path, "dest_path", ignore_patterns=["*.bin"]
    )

    # then
    arg_list = get_call_args(mock_file.get_file_client.call_args_list)
    file_paths = [a[0] for a in arg_list]
    assert "model/weights.bin" not in file_paths
    assert len(file_paths) == 2



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import base64
from concurrent.futures import ThreadPoolExecutor
import fnmatch
from functools import partial
import glob
import gzip
import json
import mimetypes
import multiprocessing
import os
import re
import shutil
import tarfile
import tempfile
import time
from typing import List, Optional
import zipfile
from pathlib import Path
from typing import Tuple
from urllib.parse import urlparse
import requests

from kserve_storage.logging import logger
from kserve_storage.storage_errors import (
    raise_storage_error,
    check_http_response,
)

MODEL_MOUNT_DIRS = "/mnt/models"

_GCS_PREFIX = "gs://"
_S3_PREFIX = "s3://"
_HDFS_PREFIX = "hdfs://"
_WEBHDFS_PREFIX = "webhdfs://"
_AZURE_BLOB_RE = [
    "https://(.+?).blob.core.windows.net/(.+)",
    "https://(.+?).z[0-9]{1,2}.blob.storage.azure.net/(.+)",
]
_AZURE_FILE_RE = [
    "https://(.+?).file.core.windows.net/(.+)",
    "https://(.+?).z[0-9]{1,2}.file.storage.azure.net/(.+)",
]
_LOCAL_PREFIX = "file://"
_URI_RE = "https?://(.+)/(.+)"
_HTTP_PREFIX = "http(s)://"
_HEADERS_SUFFIX = "-headers"
_PVC_PREFIX = "/mnt/pvc"
_HF_PREFIX = "hf://"
_GIT_RE = r"https://.+\.git"

_HDFS_SECRET_DIRECTORY = "/var/secrets/kserve-hdfscreds"
_HDFS_FILE_SECRETS = ["KERBEROS_KEYTAB", "TLS_CERT", "TLS_KEY", "TLS_CA"]

# S3 parallel download configuration
_S3_MAX_FILE_CONCURRENCY = int(os.getenv("S3_MAX_FILE_CONCURRENCY", "4"))
# Global variable for S3 resource in worker processes
_worker_s3_resource = None
# Azure async download configuration
_AZURE_MAX_FILE_CONCURRENCY = int(os.getenv("AZURE_MAX_FILE_CONCURRENCY", "4"))
_AZURE_MAX_CHUNK_CONCURRENCY = int(os.getenv("AZURE_MAX_CHUNK_CONCURRENCY", "4"))


def _should_download(
    relative_path: str,
    allow_patterns: Optional[List[str]] = None,
    ignore_patterns: Optional[List[str]] = None,
) -> bool:
    """Determine whether a file should be downloaded based on allow/ignore patterns.

    Uses fnmatch semantics consistent with huggingface_hub.snapshot_download.
    Matching is performed against the relative path of the file.

    Args:
        relative_path: The relative path of the file (e.g., "subdir/model.safetensors").
        allow_patterns: If set, only files matching at least one pattern are included.
        ignore_patterns: If set, files matching any pattern are excluded.

    Returns:
        True if the file should be downloaded, False otherwise.
    """
    if allow_patterns is not None and len(allow_patterns) > 0:
        if not any(fnmatch.fnmatch(relative_path, p) for p in allow_patterns):
            return False

    if ignore_patterns is not None and len(ignore_patterns) > 0:
        if any(fnmatch.fnmatch(relative_path, p) for p in ignore_patterns):
            return False

    return True


def _parse_patterns_from_env(env_var_name: str) -> Optional[List[str]]:
    """Parse allow/ignore patterns from an environment variable.

    Supports JSON array format: '["*.safetensors", "*.json"]'
    Falls back to comma-separated: '*.safetensors,*.json'

    Returns None if the env var is not set or empty.
    """
    value = os.environ.get(env_var_name, "").strip()
    if not value:
        return None

    try:
        parsed = json.loads(value)
        if isinstance(parsed, list):
            return [str(p).strip() for p in parsed if str(p).strip()]
        return [str(parsed).strip()] if str(parsed).strip() else None
    except (json.JSONDecodeError, TypeError):
        patterns = [p.strip() for p in value.split(",") if p.strip()]
        return patterns if patterns else None


class Storage(object):
    @staticmethod
    def download_files(
        source_uris: list[str],
        out_dirs: list[str],
        allow_patterns: Optional[List[str]] = None,
        ignore_patterns: Optional[List[str]] = None,
    ) -> list[str]:
        download_fn = partial(
            Storage.download,
            allow_patterns=allow_patterns,
            ignore_patterns=ignore_patterns,
        )
        with ThreadPoolExecutor() as executor:
            model_dirs = list(executor.map(download_fn, source_uris, out_dirs))
        return model_dirs

    @staticmethod
    def download(
        uri: str,
        out_dir: Optional[str] = None,
        allow_patterns: Optional[List[str]] = None,
        ignore_patterns: Optional[List[str]] = None,
    ) -> str:
        start = time.monotonic()
        Storage._update_with_storage_spec()

        if allow_patterns is None:
            allow_patterns = _parse_patterns_from_env("STORAGE_ALLOW_PATTERNS")
        if ignore_patterns is None:
            ignore_patterns = _parse_patterns_from_env("STORAGE_IGNORE_PATTERNS")

        logger.info("Copying contents of %s to local", uri)

        if allow_patterns:
            logger.info("Allow patterns: %s", allow_patterns)
        if ignore_patterns:
            logger.info("Ignore patterns: %s", ignore_patterns)

        if uri.startswith(_PVC_PREFIX) and not os.path.exists(uri):
            raise Exception(f"Cannot locate source uri {uri} for PVC")

        is_local = uri.startswith(_LOCAL_PREFIX) or os.path.exists(uri)
        if is_local:
            if out_dir is None:
                # noop if out_dir is not set and the path is local
                model_dir = Storage._download_local(uri)
            else:
                if not os.path.exists(out_dir):
                    os.mkdir(out_dir)
                model_dir = Storage._download_local(
                    uri, out_dir, allow_patterns, ignore_patterns
                )
        else:
            if out_dir is None:
                out_dir = tempfile.mkdtemp()
            elif not os.path.exists(out_dir):
                os.mkdir(out_dir)

            if uri.startswith(MODEL_MOUNT_DIRS):
                # Don't need to download models if this InferenceService is running in the multi-model
                # serving mode. The model agent will download models.
                model_dir = out_dir
            elif uri.startswith(_GCS_PREFIX):
                model_dir = Storage._download_gcs(
                    uri, out_dir, allow_patterns, ignore_patterns
                )
            elif uri.startswith(_S3_PREFIX):
                model_dir = Storage._download_s3(
                    uri, out_dir, allow_patterns, ignore_patterns
                )
            elif uri.startswith(_HDFS_PREFIX) or uri.startswith(_WEBHDFS_PREFIX):
                model_dir = Storage._download_hdfs(
                    uri, out_dir, allow_patterns, ignore_patterns
                )
            elif any(re.search(pattern, uri) for pattern in _AZURE_BLOB_RE):
                model_dir = Storage._download_azure_blob(
                    uri, out_dir, allow_patterns, ignore_patterns
                )
            elif any(re.search(pattern, uri) for pattern in _AZURE_FILE_RE):
                model_dir = Storage._download_azure_file_share(
                    uri, out_dir, allow_patterns, ignore_patterns
                )
            elif uri.startswith(_HF_PREFIX):
                model_dir = Storage._download_hf(
                    uri, out_dir, allow_patterns, ignore_patterns
                )
            elif re.search(_GIT_RE, uri):
                model_dir = Storage._download_git_repo(uri, out_dir)
            # "catch-all" pattern, should always be last
            elif re.search(_URI_RE, uri):
                model_dir = Storage._download_from_uri(uri, out_dir)
            else:
                raise Exception(
                    "Cannot recognize storage type for "
                    + uri
                    + "\n'%s', '%s', '%s', '%s' and '%s' are the current available storage type."
                    % (_GCS_PREFIX, _S3_PREFIX, _LOCAL_PREFIX, _HTTP_PREFIX, _HF_PREFIX)
                )

        logger.info("Successfully copied %s to %s", uri, out_dir)
        logger.info(f"Model downloaded in {time.monotonic() - start} seconds.")
        return model_dir

    @staticmethod
    def _update_with_storage_spec():
        storage_secret_json = json.loads(os.environ.get("STORAGE_CONFIG", "{}"))
        storage_secret_override_params = json.loads(
            os.environ.get("STORAGE_OVERRIDE_CONFIG", "{}")
        )
        if storage_secret_override_params:
            for key, value in storage_secret_override_params.items():
                storage_secret_json[key] = value

        if storage_secret_json.get("type", "") == "s3":
            for env_var, key in (
                ("AWS_ENDPOINT_URL", "endpoint_url"),
                ("AWS_ACCESS_KEY_ID", "access_key_id"),
                ("AWS_SECRET_ACCESS_KEY", "secret_access_key"),
                ("AWS_DEFAULT_REGION", "region"),
                ("AWS_CA_BUNDLE", "ca_bundle"),
                ("S3_VERIFY_SSL", "verify_ssl"),
                ("awsAnonymousCredential", "anonymous"),
            ):
                if key in storage_secret_json:
                    os.environ[env_var] = storage_secret_json.get(key)

        if (
            storage_secret_json.get("type", "") == "hdfs"
            or storage_secret_json.get("type", "") == "webhdfs"
        ):
            temp_dir = tempfile.mkdtemp()
            os.environ["HDFS_SECRET_DIR"] = temp_dir
            for key, value in storage_secret_json.items():
                mode = "w"

                # If the secret is supposed to be a file, then it was base64 encoded in the json
                if key in _HDFS_FILE_SECRETS:
                    value = base64.b64decode(value)
                    mode = "wb"

                with open(f"{temp_dir}/{key}", mode) as f:
                    f.write(value)
                    f.flush()

    @staticmethod
    def get_S3_config():
        from botocore import UNSIGNED
        from botocore.client import Config

        # Configurable timeouts and retries
        connect_timeout = int(os.getenv("S3_CONNECT_TIMEOUT", "15"))
        read_timeout = int(os.getenv("S3_READ_TIMEOUT", "30"))
        max_attempts = int(os.getenv("S3_MAX_ATTEMPTS", "3"))

        # default s3 config with timeouts
        c = Config(
            connect_timeout=connect_timeout,
            read_timeout=read_timeout,
            retries={"max_attempts": max_attempts, "mode": "standard"},
        )

        # anon environment variable defined in s3_secret.go
        anon = "true" == os.getenv("awsAnonymousCredential", "false").lower()
        # S3UseVirtualBucket environment variable defined in s3_secret.go
        # use virtual hosted-style URLs if enabled
        virtual = "true" == os.getenv("S3_USER_VIRTUAL_BUCKET", "false").lower()
        # S3UseAccelerate environment variable defined in s3_secret.go
        # use transfer acceleration if enabled
        accelerate = "true" == os.getenv("S3_USE_ACCELERATE", "false").lower()

        if anon:
            c = c.merge(Config(signature_version=UNSIGNED))
        if virtual:
            c = c.merge(Config(s3={"addressing_style": "virtual"}))
        if accelerate:
            c = c.merge(Config(s3={"use_accelerate_endpoint": accelerate}))

        # NOTE: If endpoint_url provided is legacy ("https://s3.amazonaws.com") and region is not global (us-east-1), set to virtual addressing style
        # So that request would not return PermanentRedirect due to region not in the endpoint url
        # AWS SDK retries under the hood to set the correct region when the valid virtual addressing style endpoint url is provided
        endpoint_url = os.getenv("AWS_ENDPOINT_URL")
        region = os.getenv("AWS_DEFAULT_REGION")
        if endpoint_url == "https://s3.amazonaws.com" and region not in (
            None,
            "us-east-1",
        ):
            c = c.merge(Config(s3={"addressing_style": "virtual"}))

        return c

    @staticmethod
    def _get_s3_client_kwargs():
        """
        Get the standardized boto3 client kwargs for S3 operations.

        Returns:
            dict: kwargs for creating boto3 S3 resource/client
        """
        kwargs = {"config": Storage.get_S3_config()}
        endpoint_url = os.getenv("AWS_ENDPOINT_URL")
        if endpoint_url:
            kwargs.update({"endpoint_url": endpoint_url})
        verify_ssl = os.getenv("S3_VERIFY_SSL")
        if verify_ssl:
            verify_ssl = verify_ssl.lower() not in ["0", "false"]
            kwargs.update({"verify": verify_ssl})
        else:
            verify_ssl = True

        # If verify_ssl is true, then check there is custom ca bundle cert
        # The CA bundle can be any local file in the container under the path
        # set in the AWS_CA_BUNDLE environment variable.
        # It can also be coming from a ConfigMap, in which case the filename
        # is cabundle.crt.
        if verify_ssl:
            global_ca_bundle_configmap = os.getenv("CA_BUNDLE_CONFIGMAP_NAME")
            isvc_aws_ca_bundle_path = os.getenv("AWS_CA_BUNDLE")
            ca_bundle_set = False
            if isvc_aws_ca_bundle_path and isvc_aws_ca_bundle_path != "":
                ca_bundle_set = True
                ca_bundle_full_path = isvc_aws_ca_bundle_path
            elif global_ca_bundle_configmap:
                ca_bundle_set = True
                global_ca_bundle_volume_mount_path = os.getenv(
                    "CA_BUNDLE_VOLUME_MOUNT_POINT"
                )
                ca_bundle_full_path = os.path.join(
                    global_ca_bundle_volume_mount_path, "cabundle.crt"
                )
            if ca_bundle_set:
                if os.path.exists(ca_bundle_full_path):
                    logger.info("ca bundle file(%s) exists." % (ca_bundle_full_path))
                    kwargs.update({"verify": ca_bundle_full_path})
                else:
                    raise RuntimeError(
                        "Failed to find ca bundle file(%s)." % ca_bundle_full_path
                    )
        return kwargs

    @staticmethod
    def _init_s3_worker():
        """
        Initialize S3 resources for worker processes.
        This runs once per worker process to avoid repeated initialization overhead.
        """
        global _worker_s3_resource
        try:
            import boto3

            kwargs = Storage._get_s3_client_kwargs()
            _worker_s3_resource = boto3.resource("s3", **kwargs)
        except Exception as e:
            logger.error(f"Failed to initialize S3 worker: {e}")
            _worker_s3_resource = None

    @staticmethod
    def _download_s3_object(args: Tuple) -> Tuple[bool, str, str]:
        """
        Worker function to download a single S3 object in a separate process.
        Uses pre-initialized S3 resource from _init_s3_worker().

        Args:
            args: Tuple containing (bucket_name:str, obj_key:str, target_path:str)

        Returns:
            Tuple of (success: bool, obj_key: str, error_message: str)
        """
        try:
            if _worker_s3_resource is None:
                return False, args[1], "S3 resource not initialized in worker process"

            bucket_name, obj_key, target_path = args
            bucket = _worker_s3_resource.Bucket(bucket_name)

            # Download the file
            bucket.download_file(obj_key, target_path)

            return True, obj_key, ""

        except Exception as e:
            return False, obj_key, str(e)

    @staticmethod
    def _download_s3(
        uri,
        temp_dir: str,
        allow_patterns: Optional[List[str]] = None,
        ignore_patterns: Optional[List[str]] = None,
    ) -> str:
        import boto3
        from botocore.exceptions import ClientError, NoCredentialsError

        # Get S3 configuration using the shared helper method
        kwargs = Storage._get_s3_client_kwargs()
        parsed = urlparse(uri, scheme="s3")
        bucket_name = parsed.netloc
        bucket_path = parsed.path.lstrip("/")

        try:
            s3 = boto3.resource("s3", **kwargs)
        except ValueError as e:
            endpoint_url = os.getenv("AWS_ENDPOINT_URL", "")
            raise_storage_error("S3", uri, e, endpoint_url)

        # Collect all objects to download
        download_tasks = []
        exact_obj_found = False

        try:
            s3.meta.client.head_bucket(Bucket=bucket_name)
            bucket = s3.Bucket(bucket_name)

            for obj in bucket.objects.filter(Prefix=bucket_path):
                if obj.key.endswith("/") or obj.size == 0:
                    logger.debug("Skipping: %s", obj.key)
                    continue

                logger.info("Found S3 object: %s (%d bytes)", obj.key, obj.size)

                if bucket_path == obj.key:
                    target_key = obj.key.rsplit("/", 1)[-1]
                    exact_obj_found = True
                else:
                    target_key = obj.key.removeprefix(bucket_path).lstrip("/")

                # Apply file filtering (skip for exact object match)
                if not exact_obj_found and not _should_download(
                    target_key, allow_patterns, ignore_patterns
                ):
                    logger.info("Skipping %s due to file pattern filter", obj.key)
                    continue

                target_path = f"{temp_dir}/{target_key}"

                # Create target directory if it doesn't exist
                if not os.path.exists(dir_path := os.path.dirname(target_path)):
                    os.makedirs(dir_path, exist_ok=True)

                download_tasks.append((bucket_name, obj.key, target_path))

                # If the exact object is found, then it is sufficient to download that and break the loop
                if exact_obj_found:
                    break
        except (ClientError, NoCredentialsError) as e:
            raise_storage_error("S3", uri, e, bucket_name)

        if len(download_tasks) == 0:
            raise RuntimeError(
                "Failed to fetch model. No model found in %s." % bucket_path
            )

        num_processes = min(_S3_MAX_FILE_CONCURRENCY, len(download_tasks))

        with multiprocessing.Pool(
            processes=num_processes, initializer=Storage._init_s3_worker
        ) as pool:
            results = list(
                pool.imap_unordered(
                    Storage._download_s3_object, download_tasks, chunksize=1
                )
            )

        # Process results and handle errors
        successful_downloads = []
        failed_downloads = []

        for success, obj_key, error_msg in results:
            if success:
                successful_downloads.append(obj_key)
                logger.info("Downloaded object %s" % obj_key)
            else:
                failed_downloads.append((obj_key, error_msg))
                logger.error("Failed to download object %s: %s" % (obj_key, error_msg))

        if len(failed_downloads) > 0:
            error_details = "; ".join(
                [f"{obj}: {err}" for obj, err in failed_downloads]
            )
            raise RuntimeError(
                f"Failed to download {len(failed_downloads)} files: {error_details}"
            )

        # Unpack compressed file, supports .tgz, tar.gz and zip file formats.
        if len(successful_downloads) == 1:
            target_path = download_tasks[0][
                2
            ]  # target_path is the 3rd element in task_args
            mimetype, _ = mimetypes.guess_type(target_path)
            if mimetype in ["application/x-tar", "application/zip"]:
                temp_dir = Storage._unpack_archive_file(target_path, mimetype, temp_dir)
        return temp_dir

    @staticmethod
    def _download_hf(
        uri,
        temp_dir: str,
        allow_patterns: Optional[List[str]] = None,
        ignore_patterns: Optional[List[str]] = None,
    ) -> str:
        from huggingface_hub import snapshot_download

        from huggingface_hub.utils import (
            RepositoryNotFoundError,
            RevisionNotFoundError,
            GatedRepoError,
            HfHubHTTPError,
        )

        components = uri[len(_HF_PREFIX) :].split("/")

        # Validate that the URI has two parts: repo and model (optional hash)
        if len(components) != 2:
            raise RuntimeError(
                "Invalid Hugging Face URI format. Expected 'hf://owner/model[:revision]', got '%s'"
                % uri
            )

        repo = components[0]
        model_part = components[1]

        if not repo:
            raise RuntimeError(
                "Hugging Face repository owner cannot be empty in URI: %s" % uri
            )
        if not model_part:
            raise RuntimeError(
                "Hugging Face model name cannot be empty in URI: %s" % uri
            )

        model, _, hash_value = model_part.partition(":")
        # Ensure model is non-empty
        if not model:
            raise RuntimeError(
                "Hugging Face model name cannot be empty in URI: %s" % uri
            )

        revision = hash_value if hash_value else None
        repo_id = f"{repo}/{model}"

        try:
            kwargs = dict(repo_id=repo_id, revision=revision, local_dir=temp_dir)
            if allow_patterns:
                kwargs["allow_patterns"] = allow_patterns
            if ignore_patterns:
                kwargs["ignore_patterns"] = ignore_patterns
            snapshot_download(**kwargs)
        except (
            RepositoryNotFoundError,
            RevisionNotFoundError,
            GatedRepoError,
            HfHubHTTPError,
        ) as e:
            raise_storage_error("HuggingFace", uri, e, repo_id)

        return temp_dir

    @staticmethod
    def _download_gcs(
        uri,
        temp_dir: str,
        allow_patterns: Optional[List[str]] = None,
        ignore_patterns: Optional[List[str]] = None,
    ) -> str:
        from google.auth import exceptions as auth_exceptions
        from google.cloud import storage
        from google.api_core import exceptions as api_exceptions
        import copy

        try:
            storage_client = storage.Client()
        except auth_exceptions.DefaultCredentialsError:
            storage_client = storage.Client.create_anonymous_client()

        bucket_args = uri.replace(_GCS_PREFIX, "", 1).split("/", 1)
        bucket_name = bucket_args[0]
        bucket_path = bucket_args[1] if len(bucket_args) > 1 else ""

        try:
            bucket = storage_client.bucket(bucket_name)
            prefix = bucket_path
            if not prefix.endswith("/"):
                prefix = prefix + "/"
            blobs = bucket.list_blobs(prefix=prefix)
            file_count = 0

            # Shallow copy, otherwise Iterator has already started
            shallow_blobs = copy.copy(blobs)
            blob = bucket.blob(bucket_path)
            # checks if the blob is a file or a directory
            if blob.name == bucket_path and len(list(shallow_blobs)) == 0:
                dest_path = os.path.join(temp_dir, os.path.basename(bucket_path))
                logger.info("Downloading single file to: %s", dest_path)
                blob.download_to_filename(dest_path)
                file_count = 1

            else:
                for blob in blobs:
                    # Replace any prefix from the object key with temp_dir
                    subdir_object_key = blob.name.replace(bucket_path, "", 1).lstrip(
                        "/"
                    )
                    # Create necessary subdirectory to store the object locally
                    if "/" in subdir_object_key:
                        local_object_dir = os.path.join(
                            temp_dir, subdir_object_key.rsplit("/", 1)[0]
                        )
                        if not os.path.isdir(local_object_dir):
                            os.makedirs(local_object_dir, exist_ok=True)
                    if (
                        subdir_object_key.strip() != ""
                        and not subdir_object_key.endswith("/")
                    ):
                        if not _should_download(
                            subdir_object_key, allow_patterns, ignore_patterns
                        ):
                            logger.info(
                                "Skipping %s due to file pattern filter", blob.name
                            )
                            continue
                        dest_path = os.path.join(temp_dir, subdir_object_key)
                        logger.info("Downloading: %s", dest_path)
                        blob.download_to_filename(dest_path)
                        file_count += 1
        except (
            api_exceptions.Forbidden,
            api_exceptions.NotFound,
            auth_exceptions.GoogleAuthError,
        ) as e:
            raise_storage_error("GCS", uri, e, bucket_name)

        if file_count == 0:
            raise RuntimeError("Failed to fetch model. No model found in %s." % uri)

        # Unpack compressed file, supports .tgz, tar.gz and zip file formats.
        if file_count == 1:
            mimetype, _ = mimetypes.guess_type(blob.name)
            if mimetype in ["application/x-tar", "application/zip"]:
                temp_dir = Storage._unpack_archive_file(dest_path, mimetype, temp_dir)
        return temp_dir

    @staticmethod
    def _load_hdfs_configuration() -> dict:
        config = {
            "HDFS_NAMENODE": None,
            "USER_PROXY": None,
            "HDFS_ROOTPATH": None,
            "TLS_CERT": None,
            "TLS_KEY": None,
            "TLS_CA": None,
            "TLS_SKIP_VERIFY": "false",
            "HEADERS": None,
            "N_THREADS": "2",
            "KERBEROS_KEYTAB": None,
            "KERBEROS_PRINCIPAL": None,
        }

        secret_dir = _HDFS_SECRET_DIRECTORY
        if os.environ.get("HDFS_SECRET_DIR"):
            secret_dir = os.environ["HDFS_SECRET_DIR"]

        for filename in os.listdir(secret_dir):
            if filename not in config:
                continue

            # We don't read files which are supposed to be files, just save their path
            if filename in _HDFS_FILE_SECRETS:
                config[filename] = f"{secret_dir}/{filename}"
                continue

            # Read file and save value in config dict
            with open(f"{secret_dir}/{filename}") as f:
                config[filename] = f.read()

        return config

    @staticmethod
    def _download_hdfs(
        uri,
        out_dir: str,
        allow_patterns: Optional[List[str]] = None,
        ignore_patterns: Optional[List[str]] = None,
    ) -> str:
        from krbcontext.context import krbContext
        from hdfs.ext.kerberos import Client, KerberosClient
        from hdfs.util import HdfsError

        config = Storage._load_hdfs_configuration()

        logger.info("Using the following hdfs config: %s", config)

        # Remove hdfs:// or webhdfs:// from the uri to get just the path
        # e.g. hdfs://user/me/model -> user/me/model
        if uri.startswith(_HDFS_PREFIX):
            path = uri[len(_HDFS_PREFIX) :]
        else:
            path = uri[len(_WEBHDFS_PREFIX) :]

        if not config["HDFS_ROOTPATH"]:
            path = "/" + path

        s = requests.Session()

        if config["TLS_CERT"]:
            s.cert = (config["TLS_CERT"], config["TLS_KEY"])
        # s.verify = , True, False, or CA PATH
        if config["TLS_CA"]:
            s.verify = config["TLS_CA"]
        if config["TLS_SKIP_VERIFY"].lower() == "true":
            s.verify = False

        if config["HEADERS"]:
            headers = json.loads(config["HEADERS"])
            s.headers.update(headers)

        try:
            if config["KERBEROS_PRINCIPAL"]:
                context = krbContext(
                    using_keytab=True,
                    principal=config["KERBEROS_PRINCIPAL"],
                    keytab_file=config["KERBEROS_KEYTAB"],
                )
                context.init_with_keytab()
                client = KerberosClient(
                    config["HDFS_NAMENODE"],
                    proxy=config["USER_PROXY"],
                    root=config["HDFS_ROOTPATH"],
                    session=s,
                )
            else:
                client = Client(
                    config["HDFS_NAMENODE"],
                    proxy=config["USER_PROXY"],
                    root=config["HDFS_ROOTPATH"],
                    session=s,
                )
            file_count = 0
            dest_file_path = ""

            # Check path exists and get path status
            # Raises HdfsError when path does not exist
            status = client.status(path)

            if status["type"] == "FILE":
                client.download(path, out_dir, n_threads=1)
                file_count += 1
                file_name = path.rsplit("/", 1)[-1]
                dest_file_path = f"{out_dir}/{file_name}"
            else:
                files = client.list(path)
                for f in files:
                    if not _should_download(f, allow_patterns, ignore_patterns):
                        logger.info("Skipping %s due to file pattern filter", f)
                        continue
                    client.download(
                        f"{path}/{f}", out_dir, n_threads=int(config["N_THREADS"])
                    )
                    file_count += 1
                    dest_file_path = f"{out_dir}/{f}"
        except (HdfsError, requests.exceptions.ConnectionError) as e:
            raise_storage_error("HDFS", uri, e, config["HDFS_NAMENODE"])

        if file_count == 1:
            mimetype, _ = mimetypes.guess_type(dest_file_path)
            if mimetype in ["application/x-tar", "application/zip"]:
                out_dir = Storage._unpack_archive_file(
                    dest_file_path, mimetype, out_dir
                )
        return out_dir

    @staticmethod
    async def _download_azure_blob_async(
        uri,
        out_dir: str,
        allow_patterns: Optional[List[str]] = None,
        ignore_patterns: Optional[List[str]] = None,
    ) -> str:
        """Async Azure blob download with chunked streaming and multi-level semaphores"""
        from azure.storage.blob.aio import BlobServiceClient
        from azure.core.exceptions import (
            ClientAuthenticationError,
            ResourceNotFoundError,
            HttpResponseError,
        )

        account_name, account_url, container_name, prefix = Storage._parse_azure_uri(
            uri
        )
        logger.info(
            "Connecting to BLOB account: [%s], container: [%s], prefix: [%s]",
            account_name,
            container_name,
            prefix,
        )

        token = (
            Storage._get_azure_storage_token()
            or Storage._get_azure_storage_access_key()
        )
        if token is None:
            logger.warning(
                "Azure credentials or shared access signature token not found, retrying anonymous access"
            )

        # File-level semaphore to control concurrent file downloads
        file_semaphore = asyncio.Semaphore(_AZURE_MAX_FILE_CONCURRENCY)

        try:
            async with BlobServiceClient(
                account_url, credential=token
            ) as blob_service_client:
                container_client = blob_service_client.get_container_client(
                    container_name
                )

                # Get all blobs using flat listing (no delimiter) to get all files regardless of hierarchy
                blobs = []
                logger.info("Listing blobs with prefix: %s", prefix)
                async for blob in container_client.list_blobs(name_starts_with=prefix):
                    logger.info("Found blob: %s (%d bytes)", blob.name, blob.size)
                    if blob.size > 0:
                        relative_path = blob.name.replace(prefix, "", 1).lstrip("/")
                        if not relative_path:
                            relative_path = os.path.basename(prefix)
                        if not _should_download(
                            relative_path, allow_patterns, ignore_patterns
                        ):
                            logger.info(
                                "Skipping %s due to file pattern filter", blob.name
                            )
                            continue
                        blobs.append(blob)

                if not blobs:
                    raise RuntimeError(
                        "Failed to fetch model. No model found in %s." % uri
                    )

                # Create download tasks with semaphore control
                download_tasks = [
                    Storage._download_single_blob_async(
                        container_client, blob, out_dir, prefix, file_semaphore
                    )
                    for blob in blobs
                ]

                # Execute all downloads concurrently
                results = await asyncio.gather(*download_tasks, return_exceptions=True)

                # Check for exceptions
                file_count = 0
                dest_path = None
                for i, result in enumerate(results):
                    if isinstance(result, Exception):
                        logger.error("Blob %s failed: %s", blobs[i].name, result)
                        raise result
                    else:
                        dest_path = result
                        file_count += 1
        except (
            ClientAuthenticationError,
            ResourceNotFoundError,
            HttpResponseError,
        ) as e:
            raise_storage_error("Azure", uri, e, container_name)

        # Handle single file unpacking
        if file_count == 1:
            mimetype, _ = mimetypes.guess_type(dest_path)
            if mimetype in ["application/x-tar", "application/zip"]:
                out_dir = Storage._unpack_archive_file(dest_path, mimetype, out_dir)

        return out_dir

    @staticmethod
    async def _download_single_blob_async(
        container_client,
        blob,
        out_dir: str,
        prefix: str,
        file_semaphore: asyncio.Semaphore,
    ) -> str:
        """Download a single blob with chunked streaming"""
        async with file_semaphore:  # Limit concurrent file downloads
            file_name = blob.name.replace(prefix, "", 1).lstrip("/")
            if not file_name:
                file_name = os.path.basename(prefix)
            dest_path = os.path.join(out_dir, file_name)
            Path(os.path.dirname(dest_path)).mkdir(parents=True, exist_ok=True)

            logger.info("Downloading: %s to %s", blob.name, dest_path)

            # Get blob client and download with chunked streaming
            blob_client = container_client.get_blob_client(blob.name)

            # Download with streaming chunks to avoid memory overload
            stream = await blob_client.download_blob(
                max_concurrency=_AZURE_MAX_CHUNK_CONCURRENCY
            )

            with open(dest_path, "wb") as f:
                # Stream chunks instead of loading entire file into memory
                async for chunk in stream.chunks():
                    f.write(chunk)

            return dest_path

    @staticmethod
    def _download_azure_file_share(
        uri,
        out_dir: str,
        allow_patterns: Optional[List[str]] = None,
        ignore_patterns: Optional[List[str]] = None,
    ) -> str:  # pylint: disable=too-many-locals
        from azure.storage.fileshare import ShareServiceClient
        from azure.core.exceptions import (
            ClientAuthenticationError,
            ResourceNotFoundError,
            HttpResponseError,
        )

        account_name, account_url, share_name, prefix = Storage._parse_azure_uri(uri)
        logger.info(
            "Connecting to file share account: [%s], container: [%s], prefix: [%s]",
            account_name,
            share_name,
            prefix,
        )
        access_key = Storage._get_azure_storage_access_key()
        if access_key is None:
            logger.warning(
                "Azure storage access key not found, retrying anonymous access"
            )

        try:
            share_service_client = ShareServiceClient(
                account_url, credential=access_key
            )
            share_client = share_service_client.get_share_client(share_name)
            file_count = 0
            share_files = []
            max_depth = 5
            stack = [(prefix, max_depth)]
            while stack:
                curr_prefix, depth = stack.pop()
                if depth < 0:
                    continue
                for item in share_client.list_directories_and_files(
                    directory_name=curr_prefix
                ):
                    if item.is_directory:
                        stack.append(
                            ("/".join([curr_prefix, item.name]).strip("/"), depth - 1)
                        )
                    else:
                        share_files.append((curr_prefix, item))
            for prefix, file_item in share_files:
                parts = [prefix] if prefix else []
                parts.append(file_item.name)
                file_path = "/".join(parts).lstrip("/")
                if not _should_download(file_path, allow_patterns, ignore_patterns):
                    logger.info("Skipping %s due to file pattern filter", file_path)
                    continue
                dest_path = os.path.join(out_dir, file_path)
                Path(os.path.dirname(dest_path)).mkdir(parents=True, exist_ok=True)
                logger.info("Downloading: %s to %s", file_item.name, dest_path)
                file_client = share_client.get_file_client(file_path)
                with open(dest_path, "wb+") as f:
                    data = file_client.download_file()
                    data.readinto(f)
                file_count += 1
        except (
            ClientAuthenticationError,
            ResourceNotFoundError,
            HttpResponseError,
        ) as e:
            raise_storage_error("Azure", uri, e, share_name)

        if file_count == 0:
            raise RuntimeError("Failed to fetch model. No model found in %s." % (uri))

        # Unpack compressed file, supports .tgz, tar.gz and zip file formats.
        if file_count == 1:
            mimetype, _ = mimetypes.guess_type(dest_path)
            if mimetype in ["application/x-tar", "application/zip"]:
                out_dir = Storage._unpack_archive_file(dest_path, mimetype, out_dir)
        return out_dir

    @staticmethod
    def _download_azure_blob(
        uri,
        out_dir: str,
        allow_patterns: Optional[List[str]] = None,
        ignore_patterns: Optional[List[str]] = None,
    ) -> str:
        """Wrapper to run async blob download"""
        return asyncio.run(
            Storage._download_azure_blob_async(
                uri, out_dir, allow_patterns, ignore_patterns
            )
        )

    @staticmethod
    def _parse_azure_uri(uri):  # pylint: disable=too-many-locals
        parsed = urlparse(uri)
        account_name = parsed.netloc.split(".")[0]
        account_url = "https://{}{}".format(
            parsed.netloc, "?" + parsed.query if parsed.query else ""
        )
        object_name, prefix = parsed.path.lstrip("/").split("/", 1)
        prefix = prefix.strip("/")
        return account_name, account_url, object_name, prefix

    @staticmethod
    def _get_azure_storage_token():
        tenant_id = os.getenv("AZ_TENANT_ID", "")
        client_id = os.getenv("AZ_CLIENT_ID", "")
        client_secret = os.getenv("AZ_CLIENT_SECRET", "")

        # convert old environment variable to conform Azure defaults
        # see azure/identity/_constants.py
        if tenant_id:
            os.environ["AZURE_TENANT_ID"] = tenant_id
        if client_id:
            os.environ["AZURE_CLIENT_ID"] = client_id
        if client_secret:
            os.environ["AZURE_CLIENT_SECRET"] = client_secret

        client_id = os.getenv("AZURE_CLIENT_ID", "")
        if not client_id:
            return None

        # note the SP must have "Storage Blob Data Owner" perms for this to work
        from azure.identity import DefaultAzureCredential

        token_credential = DefaultAzureCredential()

        logger.info("Retrieved SP token credential for client_id: %s", client_id)
        return token_credential

    @staticmethod
    def _get_azure_storage_access_key():
        return os.getenv("AZURE_STORAGE_ACCESS_KEY")

    @staticmethod
    def _download_git_repo(uri: str, out_dir: str) -> str:
        """
        Supports authentication via:
        - Username in URL: https://username@host/repo.git
        - Username from GIT_USERNAME environment variable
        - Password from GIT_PASSWORD environment variable (from Kubernetes secret)
        """
        from dulwich import porcelain
        from dulwich.errors import GitProtocolError
        from urllib.parse import urlparse, urlunparse

        logger.info("Downloading Git repository %s into %s", uri, out_dir)

        parsed = urlparse(uri)
        username = None
        clean_uri = uri

        # Extract username from URL if present (format: https://username@host/repo.git)
        # Note: If password is in URL (https://user:pass@host/repo.git), it will be ignored
        # as passwords should come from Kubernetes secrets via GIT_PASSWORD env var
        if parsed.username:
            username = parsed.username

            # Reconstruct URI without username/password
            netloc = parsed.hostname
            if parsed.port:
                netloc += f":{parsed.port}"
            clean_parsed = parsed._replace(netloc=netloc)
            clean_uri = urlunparse(clean_parsed)

        if not username:
            username = os.getenv("GIT_USERNAME")

        password = os.getenv("GIT_PASSWORD")

        try:
            clone_kwargs = {"depth": 1}
            if username:
                clone_kwargs["username"] = username
            if password:
                clone_kwargs["password"] = password

            porcelain.clone(clean_uri, out_dir, **clone_kwargs)
            logger.info("git clone successful")

        except GitProtocolError as e:
            raise_storage_error("Git", uri, e, clean_uri)
        except Exception as e:
            raise_storage_error("Git", uri, e, clean_uri)

        return out_dir

    @staticmethod
    def _download_local(
        uri,
        out_dir=None,
        allow_patterns: Optional[List[str]] = None,
        ignore_patterns: Optional[List[str]] = None,
    ) -> str:
        local_path = uri.replace(_LOCAL_PREFIX, "", 1)
        if not os.path.exists(local_path):
            raise RuntimeError("Local path %s does not exist." % (uri))

        if out_dir is None:
            return local_path
        elif not os.path.isdir(out_dir):
            os.makedirs(out_dir, exist_ok=True)

        if os.path.isdir(local_path):
            local_path = os.path.join(local_path, "*")

        file_count = 0
        for src in glob.glob(local_path):
            _, tail = os.path.split(src)
            if not _should_download(tail, allow_patterns, ignore_patterns):
                logger.info("Skipping %s due to file pattern filter", src)
                continue
            dest_path = os.path.join(out_dir, tail)
            logger.info("Linking: %s to %s", src, dest_path)
            if not os.path.exists(dest_path):
                os.symlink(src, dest_path)
            else:
                logger.info("File %s already exist", dest_path)
            file_count += 1
        if file_count == 0:
            raise RuntimeError("Failed to fetch model. No model found in %s." % (uri))
        # Unpack compressed file, supports .tgz, tar.gz and zip file formats.
        if file_count == 1:
            mimetype, _ = mimetypes.guess_type(dest_path)
            if mimetype in ["application/x-tar", "application/zip"]:
                out_dir = Storage._unpack_archive_file(dest_path, mimetype, out_dir)
        return out_dir

    @staticmethod
    def _download_from_uri(uri, out_dir=None) -> str:
        url = urlparse(uri)
        filename = os.path.basename(url.path)
        # Determine if the symbol '?' exists in the path
        if mimetypes.guess_type(url.path)[0] is None and url.query != "":
            mimetype, encoding = mimetypes.guess_type(url.query)
        else:
            mimetype, encoding = mimetypes.guess_type(url.path)
        local_path = os.path.join(out_dir, filename)

        if filename == "":
            raise RuntimeError("No filename contained in URI: %s" % (uri))

        # Get header information from host url
        headers = {}
        host_uri = url.hostname

        headers_json = os.getenv(host_uri + _HEADERS_SUFFIX, "{}")
        headers = json.loads(headers_json)

        try:
            response = requests.get(uri, stream=True, headers=headers, timeout=30)
        except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as e:
            raise_storage_error("HTTP", uri, e, host_uri)

        with response:
            check_http_response(uri, response)
            zip_content_types = (
                "application/x-zip-compressed",
                "application/zip",
                "application/zip-compressed",
            )
            if mimetype == "application/zip" and not response.headers.get(
                "Content-Type", ""
            ).startswith(zip_content_types):
                raise RuntimeError(
                    "URI: %s did not respond with any of following 'Content-Type': "
                    % uri
                    + ", ".join(zip_content_types)
                )
            tar_content_types = (
                "application/x-tar",
                "application/x-gtar",
                "application/x-gzip",
                "application/gzip",
            )
            if mimetype == "application/x-tar" and not response.headers.get(
                "Content-Type", ""
            ).startswith(tar_content_types):
                raise RuntimeError(
                    "URI: %s did not respond with any of following 'Content-Type': "
                    % uri
                    + ", ".join(tar_content_types)
                )
            if (
                mimetype != "application/zip" and mimetype != "application/x-tar"
            ) and not response.headers.get("Content-Type", "").startswith(
                "application/octet-stream"
            ):
                raise RuntimeError(
                    "URI: %s did not respond with 'Content-Type': 'application/octet-stream'"
                    % uri
                )

            if encoding == "gzip":
                stream = gzip.GzipFile(fileobj=response.raw)
                local_path = os.path.join(out_dir, f"{filename}.tar")
            else:
                stream = response.raw
            with open(local_path, "wb") as out:
                shutil.copyfileobj(stream, out)

        if mimetype in ["application/x-tar", "application/zip"]:
            out_dir = Storage._unpack_archive_file(local_path, mimetype, out_dir)
        return out_dir

    @staticmethod
    def _unpack_archive_file(file_path, mimetype, target_dir=None) -> str:
        if not target_dir:
            target_dir = os.path.dirname(file_path)

        try:
            logger.info("Unpacking: %s", file_path)
            if mimetype == "application/x-tar":
                with tarfile.open(file_path, "r", encoding="utf-8") as archive:
                    archive.extractall(target_dir, filter="data")
            else:
                with zipfile.ZipFile(file_path, "r") as archive:
                    archive.extractall(target_dir)
        except (tarfile.TarError, zipfile.BadZipfile) as e:
            raise RuntimeError(
                "Failed to unpack archive file. The file format is not valid."
            ) from e
        os.remove(file_path)
        return target_dir



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest.mock as mock
import pytest

from kserve_storage import Storage


@mock.patch("huggingface_hub.snapshot_download")
def test_download_model(mock_snapshot_download):
    uri = "hf://example.com/model:hash_value"
    repo = "example.com"
    model = "model"
    revision = "hash_value"

    Storage.download(uri)

    mock_snapshot_download.assert_called_once_with(
        repo_id=f"{repo}/{model}",
        revision=revision,
        local_dir=mock.ANY,
    )


@mock.patch("huggingface_hub.snapshot_download")
def test_download_model_with_allow_patterns(mock_snapshot_download):
    uri = "hf://example.com/model"

    Storage._download_hf(uri, "/tmp/out", allow_patterns=["*.safetensors", "*.json"])

    mock_snapshot_download.assert_called_once_with(
        repo_id="example.com/model",
        revision=None,
        local_dir="/tmp/out",
        allow_patterns=["*.safetensors", "*.json"],
    )


@mock.patch("huggingface_hub.snapshot_download")
def test_download_model_with_ignore_patterns(mock_snapshot_download):
    uri = "hf://example.com/model"

    Storage._download_hf(uri, "/tmp/out", ignore_patterns=["*.bin", "*.gguf"])

    mock_snapshot_download.assert_called_once_with(
        repo_id="example.com/model",
        revision=None,
        local_dir="/tmp/out",
        ignore_patterns=["*.bin", "*.gguf"],
    )


@mock.patch("huggingface_hub.snapshot_download")
def test_download_model_with_both_patterns(mock_snapshot_download):
    uri = "hf://example.com/model"

    Storage._download_hf(
        uri,
        "/tmp/out",
        allow_patterns=["*.json"],
        ignore_patterns=["config.json"],
    )

    mock_snapshot_download.assert_called_once_with(
        repo_id="example.com/model",
        revision=None,
        local_dir="/tmp/out",
        allow_patterns=["*.json"],
        ignore_patterns=["config.json"],
    )


@mock.patch("huggingface_hub.snapshot_download")
def test_download_model_no_patterns_omits_kwargs(mock_snapshot_download):
    uri = "hf://example.com/model"

    Storage._download_hf(uri, "/tmp/out")

    mock_snapshot_download.assert_called_once_with(
        repo_id="example.com/model",
        revision=None,
        local_dir="/tmp/out",
    )


@mock.patch("huggingface_hub.snapshot_download")
def test_download_reads_env_patterns(mock_snapshot_download):
    uri = "hf://example.com/model"

    with mock.patch.dict(
        os.environ,
        {
            "STORAGE_ALLOW_PATTERNS": '["*.safetensors"]',
            "STORAGE_IGNORE_PATTERNS": "*.bin,*.gguf",
        },
    ):
        Storage.download(uri, out_dir="/tmp/out")

    mock_snapshot_download.assert_called_once()
    call_kwargs = mock_snapshot_download.call_args[1]
    assert call_kwargs.get("allow_patterns") == ["*.safetensors"]
    assert call_kwargs.get("ignore_patterns") == ["*.bin", "*.gguf"]


@mock.patch("huggingface_hub.snapshot_download")
def test_explicit_patterns_override_env(mock_snapshot_download):
    uri = "hf://example.com/model"

    with mock.patch.dict(
        os.environ,
        {"STORAGE_ALLOW_PATTERNS": '["*.bin"]'},
    ):
        Storage.download(uri, out_dir="/tmp/out", allow_patterns=["*.safetensors"])

    call_kwargs = mock_snapshot_download.call_args[1]
    assert call_kwargs.get("allow_patterns") == ["*.safetensors"]


@mock.patch("huggingface_hub.snapshot_download")
@pytest.mark.parametrize(
    "invalid_uri, error_message",
    [
        (
            "hf://",
            "Invalid Hugging Face URI format",
        ),  # Missing repo and model
        (
            "hf://repo_only",
            "Invalid Hugging Face URI format",
        ),  # Missing model
        ("hf:///model_only", "repository owner cannot be empty"),  # Missing repo
        (
            "hf://repo/:hash_value",
            "model name cannot be empty",
        ),  # Missing model name, hash exists
    ],
)
def test_invalid_uri(mock_snapshot_download, invalid_uri, error_message):
    with pytest.raises(RuntimeError, match=error_message):
        Storage.download(invalid_uri)

    # Ensure that snapshot_download was never called
    mock_snapshot_download.assert_not_called()



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import pytest
import botocore
import tempfile
import unittest.mock as mock

from botocore.client import Config
from botocore import UNSIGNED
from kserve_storage import Storage
from kserve_storage.storage_errors import get_storage_error_message, check_http_response


class MockPool:
    """
    Mock multiprocessing.Pool that executes tasks synchronously in the same process.
    This allows boto3 mocks to work properly in tests.
    """

    def __init__(self, processes=None, initializer=None, initargs=(), **kwargs):
        self.processes = processes
        self.initializer = initializer
        self.initargs = initargs
        # Call initializer immediately if provided (simulating worker process initialization)
        if self.initializer:
            self.initializer(*self.initargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def imap_unordered(self, func, iterable, chunksize=1):
        """Execute tasks synchronously instead of in parallel processes"""
        return [func(item) for item in iterable]


# Mock multiprocessing.Pool globally for all S3 tests
mock.patch("kserve_storage.kserve_storage.multiprocessing.Pool", MockPool).start()


def create_mock_obj(path):
    mock_obj = mock.MagicMock()
    mock_obj.key = path
    mock_obj.size = 1024  # Non-zero size to avoid being skipped
    mock_obj.is_dir = False
    return mock_obj


def create_mock_boto3_bucket(mock_storage, paths):
    mock_s3_resource = mock.MagicMock()
    mock_s3_bucket = mock.MagicMock()
    mock_s3_bucket.objects.filter.return_value = [create_mock_obj(p) for p in paths]

    mock_s3_resource.Bucket.return_value = mock_s3_bucket
    mock_storage.return_value = mock_s3_resource

    return mock_s3_bucket


def get_call_args(call_args_list):
    arg_list = []
    for call in call_args_list:
        args, _ = call
        arg_list.append(args)
    return arg_list


def expected_call_args_list_single_obj(dest, path):
    return [(f"{path}".strip("/"), f'{dest}/{path.rsplit("/", 1)[-1]}'.strip("/"))]


def expected_call_args_list(parent_key, dest, paths):
    return [(f"{parent_key}/{p}".strip("/"), f"{dest}/{p}".strip("/")) for p in paths]


# pylint: disable=protected-access


@mock.patch("boto3.resource")
def test_parent_key(mock_storage):
    # given
    bucket_name = "foo"
    paths = ["models/weights.pt", "0002.h5", "a/very/long/path/config.json"]
    object_paths = ["bar/" + p for p in paths]

    # when
    mock_boto3_bucket = create_mock_boto3_bucket(mock_storage, object_paths)
    Storage._download_s3(f"s3://{bucket_name}/bar", "dest_path")

    # then
    arg_list = get_call_args(mock_boto3_bucket.download_file.call_args_list)
    assert arg_list == expected_call_args_list("bar", "dest_path", paths)

    mock_boto3_bucket.objects.filter.assert_called_with(Prefix="bar")


@mock.patch("boto3.resource")
def test_no_key(mock_storage):
    # given
    bucket_name = "foo"
    object_paths = ["models/weights.pt", "0002.h5", "a/very/long/path/config.json"]

    # when
    mock_boto3_bucket = create_mock_boto3_bucket(mock_storage, object_paths)
    Storage._download_s3(f"s3://{bucket_name}/", "dest_path")

    # then
    arg_list = get_call_args(mock_boto3_bucket.download_file.call_args_list)
    assert arg_list == expected_call_args_list("", "dest_path", object_paths)

    mock_boto3_bucket.objects.filter.assert_called_with(Prefix="")


@mock.patch("boto3.resource")
def test_storage_s3_exception(mock_resource):
    path = "s3://foo/bar"
    # Create mock client
    mock_s3_resource = mock.MagicMock()
    mock_s3_resource.Bucket.side_effect = Exception()
    mock_resource.return_value = mock_s3_resource

    with pytest.raises(Exception):
        Storage.download(path)


@mock.patch("boto3.resource")
@mock.patch("urllib3.PoolManager")
def test_no_permission_buckets(mock_connection, mock_resource):
    bad_s3_path = "s3://random/path"
    # Access private buckets without credentials
    mock_s3_resource = mock.MagicMock()
    mock_s3_bucket = mock.MagicMock()
    mock_s3_bucket.objects.filter.return_value = [mock.MagicMock()]
    mock_s3_resource.meta.client.head_bucket.side_effect = (
        botocore.exceptions.ClientError(
            {"Error": {"Code": "403", "Message": "Forbidden"}}, "HeadBucket"
        )
    )
    mock_s3_resource.Bucket.return_value = mock_s3_bucket
    mock_resource.return_value = mock_s3_resource

    with pytest.raises(RuntimeError) as exc_info:
        Storage.download(bad_s3_path)
    assert "S3" in str(exc_info.value)


@mock.patch("boto3.resource")
def test_full_name_key(mock_storage):
    # given
    bucket_name = "foo"
    object_key = "path/to/model/name.pt"

    # when
    mock_boto3_bucket = create_mock_boto3_bucket(mock_storage, [object_key])
    Storage._download_s3(f"s3://{bucket_name}/{object_key}", "dest_path")

    # then
    arg_list = get_call_args(mock_boto3_bucket.download_file.call_args_list)
    assert arg_list == expected_call_args_list_single_obj("dest_path", object_key)

    mock_boto3_bucket.objects.filter.assert_called_with(Prefix=object_key)


@mock.patch("boto3.resource")
def test_full_name_key_root_bucket_dir(mock_resource):
    # given
    bucket_name = "foo"
    object_key = "name.pt"

    # when
    mock_boto3_bucket = create_mock_boto3_bucket(mock_resource, [object_key])
    Storage._download_s3(f"s3://{bucket_name}/{object_key}", "dest_path")

    # then
    arg_list = get_call_args(mock_boto3_bucket.download_file.call_args_list)
    assert arg_list == expected_call_args_list_single_obj("dest_path", object_key)

    mock_boto3_bucket.objects.filter.assert_called_with(Prefix=object_key)


AWS_TEST_CREDENTIALS = {
    "AWS_ACCESS_KEY_ID": "testing",
    "AWS_SECRET_ACCESS_KEY": "testing",
    "AWS_SECURITY_TOKEN": "testing",
    "AWS_SESSION_TOKEN": "testing",
}


@mock.patch("boto3.resource")
def test_multikey(mock_storage):
    # given
    bucket_name = "foo"
    paths = ["b/model.bin"]
    object_paths = ["test/a/" + p for p in paths]

    # when
    mock_boto3_bucket = create_mock_boto3_bucket(mock_storage, object_paths)
    Storage._download_s3(f"s3://{bucket_name}/test/a", "dest_path")

    # then
    arg_list = get_call_args(mock_boto3_bucket.download_file.call_args_list)
    assert arg_list == expected_call_args_list("test/a", "dest_path", paths)

    mock_boto3_bucket.objects.filter.assert_called_with(Prefix="test/a")


@mock.patch("boto3.resource")
def test_files_with_no_extension(mock_storage):

    # given
    bucket_name = "foo"
    paths = ["churn-pickle", "churn-pickle-logs", "churn-pickle-report"]
    object_paths = ["test/" + p for p in paths]

    # when
    mock_boto3_bucket = create_mock_boto3_bucket(mock_storage, object_paths)
    Storage._download_s3(f"s3://{bucket_name}/test/churn-pickle", "dest_path")

    # then
    arg_list = get_call_args(mock_boto3_bucket.download_file.call_args_list)

    # Download only the exact file if found; otherwise, download all files with the given prefix
    assert arg_list[0] == expected_call_args_list("test", "dest_path", paths)[0]

    mock_boto3_bucket.objects.filter.assert_called_with(Prefix="test/churn-pickle")


def test_get_S3_config():
    # Default config now includes configurable timeouts and retries
    DEFAULT_CONFIG = Config(
        connect_timeout=15,
        read_timeout=30,
        retries={"max_attempts": 3, "mode": "standard"},
    )
    ANON_CONFIG = Config(signature_version=UNSIGNED)
    VIRTUAL_CONFIG = Config(s3={"addressing_style": "virtual"})
    USE_ACCELERATE_CONFIG = Config(s3={"use_accelerate_endpoint": True})

    with mock.patch.dict(os.environ, {}, clear=True):
        config1 = Storage.get_S3_config()
    assert config1.connect_timeout == DEFAULT_CONFIG.connect_timeout
    assert config1.read_timeout == DEFAULT_CONFIG.read_timeout
    assert config1.retries == DEFAULT_CONFIG.retries

    with mock.patch.dict(os.environ, {"awsAnonymousCredential": "False"}, clear=True):
        config2 = Storage.get_S3_config()
    assert config2.connect_timeout == DEFAULT_CONFIG.connect_timeout

    with mock.patch.dict(os.environ, AWS_TEST_CREDENTIALS, clear=True):
        config3 = Storage.get_S3_config()
    assert config3.connect_timeout == DEFAULT_CONFIG.connect_timeout

    with mock.patch.dict(os.environ, {"awsAnonymousCredential": "True"}):
        config4 = Storage.get_S3_config()
    assert config4.signature_version == ANON_CONFIG.signature_version

    # assuming Python 3.5 or greater for joining dictionaries
    credentials_and_anon = {**AWS_TEST_CREDENTIALS, "awsAnonymousCredential": "True"}
    with mock.patch.dict(os.environ, credentials_and_anon):
        config5 = Storage.get_S3_config()
    assert config5.signature_version == ANON_CONFIG.signature_version

    with mock.patch.dict(os.environ, {"S3_USER_VIRTUAL_BUCKET": "False"}, clear=True):
        config6 = Storage.get_S3_config()
    assert config6.connect_timeout == DEFAULT_CONFIG.connect_timeout

    with mock.patch.dict(os.environ, {"S3_USER_VIRTUAL_BUCKET": "True"}, clear=True):
        config7 = Storage.get_S3_config()
    assert config7.s3["addressing_style"] == VIRTUAL_CONFIG.s3["addressing_style"]

    with mock.patch.dict(os.environ, {"S3_USE_ACCELERATE": "False"}, clear=True):
        config6 = Storage.get_S3_config()
    assert config6.connect_timeout == DEFAULT_CONFIG.connect_timeout

    with mock.patch.dict(os.environ, {"S3_USE_ACCELERATE": "True"}):
        config7 = Storage.get_S3_config()
    assert (
        config7.s3["use_accelerate_endpoint"]
        == USE_ACCELERATE_CONFIG.s3["use_accelerate_endpoint"]
    )

    # tests legacy endpoint url
    with mock.patch.dict(
        os.environ,
        {
            "AWS_ENDPOINT_URL": "https://s3.amazonaws.com",
            "AWS_DEFAULT_REGION": "eu-west-1",
        },
    ):
        config8 = Storage.get_S3_config()
    assert config8.s3["addressing_style"] == VIRTUAL_CONFIG.s3["addressing_style"]

    # tests custom timeout configuration
    with mock.patch.dict(
        os.environ,
        {
            "S3_CONNECT_TIMEOUT": "5",
            "S3_READ_TIMEOUT": "10",
            "S3_MAX_ATTEMPTS": "5",
        },
        clear=True,
    ):
        config9 = Storage.get_S3_config()
    assert config9.connect_timeout == 5
    assert config9.read_timeout == 10
    assert config9.retries["max_attempts"] == 5


def test_update_with_storage_spec_s3(monkeypatch):
    # save the environment and restore it after the test to avoid mutating it
    # since _update_with_storage_spec modifies it
    previous_env = os.environ.copy()

    monkeypatch.setenv("STORAGE_CONFIG", '{"type": "s3"}')
    Storage._update_with_storage_spec()

    for var in (
        "AWS_ENDPOINT_URL",
        "AWS_ACCESS_KEY_ID",
        "AWS_SECRET_ACCESS_KEY",
        "AWS_DEFAULT_REGION",
        "AWS_CA_BUNDLE",
        "S3_VERIFY_SSL",
        "awsAnonymousCredential",
    ):
        assert os.getenv(var) is None

    storage_config = {
        "access_key_id": "xxxxxxxxxxxxxxxxxxxx",
        "bucket": "abucketname",
        "default_bucket": "abucketname",
        "endpoint_url": "https://s3.us-east-2.amazonaws.com/",
        "region": "us-east-2",
        "secret_access_key": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        "type": "s3",
        "ca_bundle": "/path/to/cabundle.crt",
        "verify_ssl": "false",
        "anonymous": "True",
    }

    monkeypatch.setenv("STORAGE_CONFIG", json.dumps(storage_config))
    Storage._update_with_storage_spec()

    assert os.getenv("AWS_ENDPOINT_URL") == storage_config["endpoint_url"]
    assert os.getenv("AWS_ACCESS_KEY_ID") == storage_config["access_key_id"]
    assert os.getenv("AWS_SECRET_ACCESS_KEY") == storage_config["secret_access_key"]
    assert os.getenv("AWS_DEFAULT_REGION") == storage_config["region"]
    assert os.getenv("AWS_CA_BUNDLE") == storage_config["ca_bundle"]
    assert os.getenv("S3_VERIFY_SSL") == storage_config["verify_ssl"]
    assert os.getenv("awsAnonymousCredential") == storage_config["anonymous"]

    # revert changes
    os.environ.clear()
    os.environ.update(previous_env)


@mock.patch("boto3.resource")
def test_target_startswith_parent_folder_name(mock_storage):
    bucket_name = "foo"
    paths = ["model.pkl", "a/model.pkl", "conda.yaml"]
    object_paths = ["test/artifacts/model/" + p for p in paths]

    # when
    mock_boto3_bucket = create_mock_boto3_bucket(mock_storage, object_paths)
    Storage._download_s3(f"s3://{bucket_name}/test/artifacts/model", "dest_path")

    # then
    arg_list = get_call_args(mock_boto3_bucket.download_file.call_args_list)
    assert (
        arg_list[0]
        == expected_call_args_list("test/artifacts/model", "dest_path", paths)[0]
    )
    mock_boto3_bucket.objects.filter.assert_called_with(Prefix="test/artifacts/model")


@mock.patch("boto3.resource")
def test_file_name_preservation(mock_storage):
    # given
    bucket_name = "local-model"
    paths = ["MLmodel"]
    object_paths = ["model/" + p for p in paths]
    expected_file_name = "MLmodel"  # Expected file name after download

    # when
    mock_boto3_bucket = create_mock_boto3_bucket(mock_storage, object_paths)
    Storage._download_s3(f"s3://{bucket_name}/model", "dest_path")

    # then
    arg_list = get_call_args(mock_boto3_bucket.download_file.call_args_list)
    assert len(arg_list) == 1  # Ensure only one file was downloaded
    downloaded_source, downloaded_target = arg_list[0]

    # Check if the source S3 key matches the original object key
    assert (
        downloaded_source == object_paths[0]
    ), f"Expected {object_paths[0]}, got {downloaded_source}"

    # Check if the target file path ends with the expected file name
    assert downloaded_target.endswith(
        expected_file_name
    ), f"Expected file name to end with {expected_file_name}, got {downloaded_target}"

    mock_boto3_bucket.objects.filter.assert_called_with(Prefix="model")


@mock.patch("boto3.resource")
def test_target_download_path_and_name(mock_storage):
    bucket_name = "foo"
    paths = ["model.pkl", "a/model.pkl", "conda.yaml"]
    object_paths = ["model/" + p for p in paths]

    # when
    mock_boto3_bucket = create_mock_boto3_bucket(mock_storage, object_paths)
    Storage._download_s3(f"s3://{bucket_name}/model", "dest_path")

    # then
    arg_list = get_call_args(mock_boto3_bucket.download_file.call_args_list)
    assert arg_list[0] == expected_call_args_list("model", "dest_path", paths)[0]
    assert arg_list[1] == expected_call_args_list("model", "dest_path", paths)[1]

    mock_boto3_bucket.objects.filter.assert_called_with(Prefix="model")


@mock.patch("boto3.resource")
def test_ca_bundle_with_aws_ca_bundle_only(mock_storage):
    """Test that AWS_CA_BUNDLE can be used independently without CA_BUNDLE_CONFIGMAP_NAME"""
    bucket_name = "foo"
    object_key = "model.pkl"

    # Create a temporary CA bundle file
    with tempfile.NamedTemporaryFile(
        mode="w", delete=False, suffix=".crt"
    ) as temp_ca_file:
        temp_ca_file.write(
            "-----BEGIN CERTIFICATE-----\ntest\n-----END CERTIFICATE-----"
        )
        ca_bundle_path = temp_ca_file.name

    try:
        # Mock the boto3 resource and bucket
        create_mock_boto3_bucket(mock_storage, [object_key])

        # Set only AWS_CA_BUNDLE environment variable (no CA_BUNDLE_CONFIGMAP_NAME)
        with mock.patch.dict(
            os.environ,
            {"AWS_CA_BUNDLE": ca_bundle_path, "S3_VERIFY_SSL": "true"},
            clear=True,
        ):
            Storage._download_s3(f"s3://{bucket_name}/{object_key}", "dest_path")

        # Verify that boto3.resource was called twice (once for listing, once for download)
        # Both calls should have the correct verify parameter
        assert mock_storage.call_count == 2
        for call_args in mock_storage.call_args_list:
            assert call_args[1]["verify"] == ca_bundle_path

    finally:
        # Clean up the temporary file
        os.unlink(ca_bundle_path)


@mock.patch("boto3.resource")
def test_ca_bundle_with_configmap_only(mock_storage):
    """Test that CA bundle works with ConfigMap when AWS_CA_BUNDLE is not set"""
    bucket_name = "foo"
    object_key = "model.pkl"

    # Create a temporary CA bundle file in the expected ConfigMap location
    with tempfile.TemporaryDirectory() as temp_dir:
        ca_bundle_path = os.path.join(temp_dir, "cabundle.crt")
        with open(ca_bundle_path, "w") as f:
            f.write("-----BEGIN CERTIFICATE-----\ntest\n-----END CERTIFICATE-----")

        # Mock the boto3 resource and bucket
        create_mock_boto3_bucket(mock_storage, [object_key])

        # Set ConfigMap environment variables only
        with mock.patch.dict(
            os.environ,
            {
                "CA_BUNDLE_CONFIGMAP_NAME": "test-configmap",
                "CA_BUNDLE_VOLUME_MOUNT_POINT": temp_dir,
                "S3_VERIFY_SSL": "true",
            },
            clear=True,
        ):
            Storage._download_s3(f"s3://{bucket_name}/{object_key}", "dest_path")

        # Verify that boto3.resource was called twice (once for listing, once for download)
        # Both calls should have the correct verify parameter
        assert mock_storage.call_count == 2
        for call_args in mock_storage.call_args_list:
            assert call_args[1]["verify"] == ca_bundle_path


@mock.patch("boto3.resource")
def test_ca_bundle_aws_ca_bundle_takes_precedence(mock_storage):
    """Test that AWS_CA_BUNDLE takes precedence over ConfigMap when both are set"""
    bucket_name = "foo"
    object_key = "model.pkl"

    # Create temporary CA bundle files
    with tempfile.NamedTemporaryFile(
        mode="w", delete=False, suffix=".crt"
    ) as aws_ca_file:
        aws_ca_file.write(
            "-----BEGIN CERTIFICATE-----\naws_ca\n-----END CERTIFICATE-----"
        )
        aws_ca_bundle_path = aws_ca_file.name

    with tempfile.TemporaryDirectory() as temp_dir:
        configmap_ca_bundle_path = os.path.join(temp_dir, "cabundle.crt")
        with open(configmap_ca_bundle_path, "w") as f:
            f.write(
                "-----BEGIN CERTIFICATE-----\nconfigmap_ca\n-----END CERTIFICATE-----"
            )

        try:
            # Mock the boto3 resource and bucket
            create_mock_boto3_bucket(mock_storage, [object_key])

            # Set both AWS_CA_BUNDLE and ConfigMap environment variables
            with mock.patch.dict(
                os.environ,
                {
                    "AWS_CA_BUNDLE": aws_ca_bundle_path,
                    "CA_BUNDLE_CONFIGMAP_NAME": "test-configmap",
                    "CA_BUNDLE_VOLUME_MOUNT_POINT": temp_dir,
                    "S3_VERIFY_SSL": "true",
                },
                clear=True,
            ):
                Storage._download_s3(f"s3://{bucket_name}/{object_key}", "dest_path")

            # Verify that boto3.resource was called twice with AWS_CA_BUNDLE path (takes precedence)
            assert mock_storage.call_count == 2
            for call_args in mock_storage.call_args_list:
                assert call_args[1]["verify"] == aws_ca_bundle_path

        finally:
            # Clean up the temporary file
            os.unlink(aws_ca_bundle_path)


@mock.patch("boto3.resource")
def test_ca_bundle_file_not_found_aws_ca_bundle(mock_storage):
    """Test that RuntimeError is raised when AWS_CA_BUNDLE file doesn't exist"""
    bucket_name = "foo"
    object_key = "model.pkl"
    non_existent_path = "/tmp/non_existent_ca_bundle.crt"

    # Mock the boto3 resource and bucket
    create_mock_boto3_bucket(mock_storage, [object_key])

    # Set AWS_CA_BUNDLE to a non-existent file
    with mock.patch.dict(
        os.environ,
        {"AWS_CA_BUNDLE": non_existent_path, "S3_VERIFY_SSL": "true"},
        clear=True,
    ):
        with pytest.raises(
            RuntimeError,
            match=f"Failed to find ca bundle file\\({non_existent_path}\\)",
        ):
            Storage._download_s3(f"s3://{bucket_name}/{object_key}", "dest_path")


@mock.patch("boto3.resource")
def test_ca_bundle_file_not_found_configmap(mock_storage):
    """Test that RuntimeError is raised when ConfigMap CA bundle file doesn't exist"""
    bucket_name = "foo"
    object_key = "model.pkl"

    # Mock the boto3 resource and bucket
    create_mock_boto3_bucket(mock_storage, [object_key])

    # Set ConfigMap environment variables with non-existent directory
    with mock.patch.dict(
        os.environ,
        {
            "CA_BUNDLE_CONFIGMAP_NAME": "test-configmap",
            "CA_BUNDLE_VOLUME_MOUNT_POINT": "/tmp/non_existent_dir",
            "S3_VERIFY_SSL": "true",
        },
        clear=True,
    ):
        expected_path = "/tmp/non_existent_dir/cabundle.crt"
        with pytest.raises(
            RuntimeError, match=f"Failed to find ca bundle file\\({expected_path}\\)"
        ):
            Storage._download_s3(f"s3://{bucket_name}/{object_key}", "dest_path")


@mock.patch("boto3.resource")
def test_ca_bundle_verify_ssl_false_no_ca_bundle_check(mock_storage):
    """Test that CA bundle is not checked when S3_VERIFY_SSL is false"""
    bucket_name = "foo"
    object_key = "model.pkl"
    non_existent_path = "/tmp/non_existent_ca_bundle.crt"

    # Mock the boto3 resource and bucket
    create_mock_boto3_bucket(mock_storage, [object_key])

    # Set AWS_CA_BUNDLE to a non-existent file but disable SSL verification
    with mock.patch.dict(
        os.environ,
        {"AWS_CA_BUNDLE": non_existent_path, "S3_VERIFY_SSL": "false"},
        clear=True,
    ):
        # This should not raise an error because verify_ssl is False
        Storage._download_s3(f"s3://{bucket_name}/{object_key}", "dest_path")

    # Verify that boto3.resource was called twice with verify=False
    assert mock_storage.call_count == 2
    for call_args in mock_storage.call_args_list:
        assert call_args[1]["verify"] is False


@mock.patch("boto3.resource")
def test_ca_bundle_empty_aws_ca_bundle_uses_configmap(mock_storage):
    """Test that empty AWS_CA_BUNDLE falls back to ConfigMap"""
    bucket_name = "foo"
    object_key = "model.pkl"

    # Create a temporary CA bundle file in the expected ConfigMap location
    with tempfile.TemporaryDirectory() as temp_dir:
        ca_bundle_path = os.path.join(temp_dir, "cabundle.crt")
        with open(ca_bundle_path, "w") as f:
            f.write("-----BEGIN CERTIFICATE-----\ntest\n-----END CERTIFICATE-----")

        # Mock the boto3 resource and bucket
        create_mock_boto3_bucket(mock_storage, [object_key])

        # Set empty AWS_CA_BUNDLE and ConfigMap environment variables
        with mock.patch.dict(
            os.environ,
            {
                "AWS_CA_BUNDLE": "",
                "CA_BUNDLE_CONFIGMAP_NAME": "test-configmap",
                "CA_BUNDLE_VOLUME_MOUNT_POINT": temp_dir,
                "S3_VERIFY_SSL": "true",
            },
            clear=True,
        ):
            Storage._download_s3(f"s3://{bucket_name}/{object_key}", "dest_path")

        # Verify that boto3.resource was called twice with the ConfigMap path
        assert mock_storage.call_count == 2
        for call_args in mock_storage.call_args_list:
            assert call_args[1]["verify"] == ca_bundle_path


def test_error_message_no_credentials():
    from botocore.exceptions import NoCredentialsError

    error = NoCredentialsError()
    msg = get_storage_error_message("S3", error, "test-bucket")
    assert "authentication failed" in msg.lower()
    assert "AWS_ACCESS_KEY_ID" in msg or "S3" in msg


def test_error_message_access_denied():
    error = botocore.exceptions.ClientError(
        {"Error": {"Code": "AccessDenied", "Message": "Access Denied"}},
        "GetObject",
    )
    msg = get_storage_error_message("S3", error, "test-bucket")
    assert "access denied" in msg.lower() or "403" in msg


def test_error_message_bucket_not_found():
    error = botocore.exceptions.ClientError(
        {
            "Error": {
                "Code": "NoSuchBucket",
                "Message": "The specified bucket does not exist",
            }
        },
        "HeadBucket",
    )
    msg = get_storage_error_message("S3", error, "nonexistent-bucket")
    assert "does not exist" in msg.lower() or "nonexistent-bucket" in msg


def test_error_message_invalid_credentials():
    error = botocore.exceptions.ClientError(
        {
            "Error": {
                "Code": "InvalidAccessKeyId",
                "Message": "The AWS Access Key Id you provided does not exist",
            }
        },
        "HeadBucket",
    )
    msg = get_storage_error_message("S3", error, "test-bucket")
    assert "S3" in msg or "AWS" in msg or "credentials" in msg.lower()


def test_error_message_connection_error():
    error = Exception("connection timeout")
    msg = get_storage_error_message("S3", error, "endpoint.example.com")
    assert "connect" in msg.lower() or "timeout" in msg.lower()


def test_error_message_not_found():
    error = Exception("resource not found")
    msg = get_storage_error_message("GCS", error, "my-bucket")
    assert "not found" in msg.lower()


def test_error_message_gcs_auth():
    error = Exception("credentials error")
    msg = get_storage_error_message("GCS", error, "my-bucket")
    assert "authentication failed" in msg.lower() or "GCS" in msg


def test_error_message_azure_auth():
    error = Exception("authentication failed")
    msg = get_storage_error_message("Azure", error, "my-container")
    assert "authentication failed" in msg.lower()


def test_error_message_huggingface():
    error = Exception("gated repo access denied")
    msg = get_storage_error_message("HuggingFace", error, "org/model")
    assert "HuggingFace" in msg or "HF_TOKEN" in msg


def test_error_message_default():
    error = Exception("some unknown error")
    msg = get_storage_error_message("S3", error, "test-bucket")
    assert "S3" in msg
    assert "unknown error" in msg.lower()


@pytest.mark.parametrize(
    "status_code, expected_message",
    [
        (401, "authentication"),
        (403, "access denied"),
        (404, "not found"),
        (500, "500"),
    ],
)
def test_check_http_response_errors(status_code, expected_message):
    mock_response = mock.MagicMock()
    mock_response.status_code = status_code
    with pytest.raises(RuntimeError) as exc_info:
        check_http_response("https://example.com/model", mock_response)
    assert expected_message in str(exc_info.value).lower()


def test_check_http_response_200_ok():
    mock_response = mock.MagicMock()
    mock_response.status_code = 200
    check_http_response("https://example.com/model", mock_response)


@mock.patch("boto3.resource")
def test_s3_no_credentials_error(mock_resource):
    from botocore.exceptions import NoCredentialsError

    mock_s3_resource = mock.MagicMock()
    mock_s3_resource.meta.client.head_bucket.side_effect = NoCredentialsError()
    mock_resource.return_value = mock_s3_resource

    with pytest.raises(RuntimeError) as exc_info:
        Storage._download_s3("s3://test-bucket/model", "/tmp/dest")
    assert "authentication" in str(exc_info.value).lower()


@mock.patch("boto3.resource")
def test_s3_invalid_endpoint_error(mock_resource):
    mock_resource.side_effect = ValueError("Invalid endpoint: bad-url")

    with pytest.raises(RuntimeError) as exc_info:
        Storage._download_s3("s3://test-bucket/model", "/tmp/dest")
    assert "S3" in str(exc_info.value)


@mock.patch("boto3.resource")
def test_s3_allow_patterns(mock_storage):
    bucket_name = "foo"
    object_paths = [
        "model/weights.safetensors",
        "model/weights.bin",
        "model/config.json",
    ]
    mock_boto3_bucket = create_mock_boto3_bucket(mock_storage, object_paths)

    Storage._download_s3(
        f"s3://{bucket_name}/model",
        "dest_path",
        allow_patterns=["*.safetensors", "*.json"],
    )

    arg_list = get_call_args(mock_boto3_bucket.download_file.call_args_list)
    downloaded_keys = [a[0] for a in arg_list]
    assert "model/weights.safetensors" in downloaded_keys
    assert "model/config.json" in downloaded_keys
    assert "model/weights.bin" not in downloaded_keys


@mock.patch("boto3.resource")
def test_s3_ignore_patterns(mock_storage):
    bucket_name = "foo"
    object_paths = [
        "model/weights.safetensors",
        "model/weights.bin",
        "model/config.json",
    ]
    mock_boto3_bucket = create_mock_boto3_bucket(mock_storage, object_paths)

    Storage._download_s3(
        f"s3://{bucket_name}/model",
        "dest_path",
        ignore_patterns=["*.bin"],
    )

    arg_list = get_call_args(mock_boto3_bucket.download_file.call_args_list)
    downloaded_keys = [a[0] for a in arg_list]
    assert "model/weights.bin" not in downloaded_keys
    assert len(downloaded_keys) == 2


@mock.patch("boto3.resource")
def test_s3_exact_object_bypasses_filter(mock_storage):
    """When a URI points to an exact object, filtering should not apply."""
    bucket_name = "foo"
    object_key = "model/weights.bin"
    mock_boto3_bucket = create_mock_boto3_bucket(mock_storage, [object_key])

    # Even with allow_patterns that exclude .bin, exact object should still download
    Storage._download_s3(
        f"s3://{bucket_name}/{object_key}",
        "dest_path",
        allow_patterns=["*.safetensors"],
    )

    arg_list = get_call_args(mock_boto3_bucket.download_file.call_args_list)
    assert len(arg_list) == 1
    assert arg_list[0][0] == object_key


@mock.patch("boto3.resource")
def test_s3_no_patterns_downloads_all(mock_storage):
    bucket_name = "foo"
    object_paths = [
        "model/weights.safetensors",
        "model/weights.bin",
        "model/config.json",
    ]
    mock_boto3_bucket = create_mock_boto3_bucket(mock_storage, object_paths)

    Storage._download_s3(f"s3://{bucket_name}/model", "dest_path")

    arg_list = get_call_args(mock_boto3_bucket.download_file.call_args_list)
    assert len(arg_list) == 3



# Copyright 2024 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest.mock as mock
import pytest
from kserve_storage import Storage

STORAGE_MODULE = "kserve_storage.kserve_storage"


def get_call_args(call_args_list):
    arg_list = []
    for call in call_args_list:
        args, _ = call
        arg_list.append(args)
    return arg_list


def create_mock_dir(name):
    mock_dir = mock.MagicMock()
    mock_dir.name = name
    return mock_dir


def create_mock_dir_with_file(dir_name, file_name):
    mock_obj = mock.MagicMock()
    mock_obj.name = f"{dir_name}/{file_name}"
    return mock_obj


@mock.patch("google.cloud.storage.Client")
def test_gcs_with_empty_dir(mock_client):
    gcs_path = "gs://foo/bar"
    mock_bucket = mock.MagicMock()

    mock_bucket.list_blobs().__iter__.return_value = [create_mock_dir("bar/")]
    mock_client.return_value.bucket.return_value = mock_bucket

    with pytest.raises(Exception):
        Storage.download(gcs_path)


@mock.patch("google.cloud.storage.Client")
def test_mock_gcs(mock_client):
    gcs_path = "gs://foo/bar"
    mock_root_dir = create_mock_dir("bar/")
    mock_file = create_mock_dir_with_file("bar", "mock.object")

    mock_bucket = mock.MagicMock()
    mock_bucket.list_blobs().__iter__.return_value = [mock_root_dir, mock_file]
    mock_client.return_value.bucket.return_value = mock_bucket
    assert Storage.download(gcs_path)


@mock.patch("google.cloud.storage.Client")
def test_gcs_with_nested_sub_dir(mock_client):
    gcs_path = "gs://foo/bar/test"

    mock_root_dir = create_mock_dir("bar/")
    mock_sub_dir = create_mock_dir("test/")
    mock_file = create_mock_dir_with_file("test", "mock.object")

    mock_bucket = mock.MagicMock()
    mock_bucket.list_blobs().__iter__.return_value = [
        mock_root_dir,
        mock_sub_dir,
        mock_file,
    ]
    mock_client.return_value.bucket.return_value = mock_bucket

    Storage.download(gcs_path)

    arg_list = get_call_args(mock_file.download_to_filename.call_args_list)
    assert "test/mock.object" in arg_list[0][0]


@mock.patch("google.cloud.storage.Client")
def test_download_model_from_gcs(mock_client):
    gcs_path = "gs://foo/bar"

    mock_dir = create_mock_dir("bar/")
    mock_file = create_mock_dir_with_file("bar", "mock.object")

    mock_bucket = mock.MagicMock()
    mock_bucket.list_blobs().__iter__.return_value = [
        mock_dir,
        mock_file,
    ]
    mock_client.return_value.bucket.return_value = mock_bucket

    Storage.download(gcs_path)

    arg_list = get_call_args(mock_file.download_to_filename.call_args_list)
    assert "/mock.object" in arg_list[0][0]


@mock.patch("google.cloud.storage.Client")
def test_download_model_from_gcs_as_single_file(mock_client):
    gcs_path = "gs://foo/bar/mock.object"
    mock_file = create_mock_dir_with_file("bar", "mock.object")

    mock_bucket = mock.MagicMock()
    mock_bucket.blob.return_value = mock_file
    mock_file.exists.return_value = True
    mock_client.return_value.bucket.return_value = mock_bucket

    Storage.download(gcs_path)
    arg_list = get_call_args(mock_file.download_to_filename.call_args_list)

    assert "/mock.object" in arg_list[0][0]


@mock.patch("os.remove")
@mock.patch("os.mkdir")
@mock.patch("zipfile.ZipFile")
@mock.patch("google.cloud.storage.Client")
def test_gcs_model_unpack_archive_file(
    mock_client, MockZipFile, mock_create, mock_remove
):
    gcs_path = "gs://foo/bar"
    output_dir = "test/out_dir"

    mock_dir = create_mock_dir("bar/")
    mock_file = create_mock_dir_with_file("bar", "mock.zip")

    # Properly set up the mock zipfile context manager
    mock_zipfile = mock.MagicMock()
    MockZipFile.return_value = mock_zipfile
    mock_zipfile.__enter__.return_value = mock_zipfile

    mock_bucket = mock.MagicMock()
    mock_bucket.list_blobs().__iter__.return_value = [
        mock_dir,
        mock_file,
    ]
    mock_client.return_value.bucket.return_value = mock_bucket

    Storage.download(gcs_path, output_dir)

    download_arg_list = get_call_args(mock_file.download_to_filename.call_args_list)

    # Check that extractall was called on the zipfile mock
    assert len(download_arg_list) > 0, "download_to_filename was never called"
    assert "/mock.zip" in download_arg_list[0][0]
    assert mock_zipfile.extractall.called
    assert output_dir == mock_zipfile.extractall.call_args[0][0]
    assert mock_remove.called


@mock.patch("google.cloud.storage.Client")
def test_gcs_allow_patterns(mock_client):
    gcs_path = "gs://foo/bar"

    mock_dir = create_mock_dir("bar/")
    mock_safetensors = create_mock_dir_with_file("bar", "weights.safetensors")
    mock_bin = create_mock_dir_with_file("bar", "weights.bin")
    mock_json = create_mock_dir_with_file("bar", "config.json")

    mock_bucket = mock.MagicMock()
    mock_bucket.list_blobs().__iter__.return_value = [
        mock_dir,
        mock_safetensors,
        mock_bin,
        mock_json,
    ]
    mock_client.return_value.bucket.return_value = mock_bucket

    Storage._download_gcs(
        gcs_path, "/tmp/dest", allow_patterns=["*.safetensors", "*.json"]
    )

    # safetensors and json should be downloaded, bin should not
    assert mock_safetensors.download_to_filename.called
    assert mock_json.download_to_filename.called
    assert not mock_bin.download_to_filename.called


@mock.patch("google.cloud.storage.Client")
def test_gcs_ignore_patterns(mock_client):
    gcs_path = "gs://foo/bar"

    mock_dir = create_mock_dir("bar/")
    mock_safetensors = create_mock_dir_with_file("bar", "weights.safetensors")
    mock_bin = create_mock_dir_with_file("bar", "weights.bin")
    mock_json = create_mock_dir_with_file("bar", "config.json")

    mock_bucket = mock.MagicMock()
    mock_bucket.list_blobs().__iter__.return_value = [
        mock_dir,
        mock_safetensors,
        mock_bin,
        mock_json,
    ]
    mock_client.return_value.bucket.return_value = mock_bucket

    Storage._download_gcs(gcs_path, "/tmp/dest", ignore_patterns=["*.bin"])

    # bin should not be downloaded, others should
    assert mock_safetensors.download_to_filename.called
    assert mock_json.download_to_filename.called
    assert not mock_bin.download_to_filename.called


@mock.patch("google.cloud.storage.Client")
def test_gcs_no_patterns_downloads_all(mock_client):
    gcs_path = "gs://foo/bar"

    mock_dir = create_mock_dir("bar/")
    mock_safetensors = create_mock_dir_with_file("bar", "weights.safetensors")
    mock_bin = create_mock_dir_with_file("bar", "weights.bin")
    mock_json = create_mock_dir_with_file("bar", "config.json")

    mock_bucket = mock.MagicMock()
    mock_bucket.list_blobs().__iter__.return_value = [
        mock_dir,
        mock_safetensors,
        mock_bin,
        mock_json,
    ]
    mock_client.return_value.bucket.return_value = mock_bucket

    Storage._download_gcs(gcs_path, "/tmp/dest")

    # all files should be downloaded
    assert mock_safetensors.download_to_filename.called
    assert mock_bin.download_to_filename.called
    assert mock_json.download_to_filename.called
